/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include <draw/draw.pmh>

namespace fe
{

ViewCommon::ViewCommon()
{
	m_spCameraDisplay=new CameraDisplay();
	m_spCameraDisplay->setName("ViewCommon.CameraDisplay");
}

void ViewCommon::screenAlignment(SpatialTransform &a_rotation)
{
	SpatialVector zero(0.0,0.0,0.0);
	SpatialVector v[3];
	v[0] = unproject(1000.0, 0.0, 0.0, CameraI::e_current);
	v[1] = unproject(0.0, 1000.0, 0.0, CameraI::e_current);
	v[2] = unproject(0.0, 0.0, 0.0, CameraI::e_current);
	v[0] -= v[2];
	v[1] -= v[2];
	cross3(v[2], v[1], v[0]);
	cross3(v[0], v[1], v[2]);
	if(v[0] == zero || v[1] == zero || v[2] == zero)
	{
		v[0] = SpatialVector(1.0,0.0,0.0);
		v[1] = SpatialVector(0.0,1.0,0.0);
		v[2] = SpatialVector(0.0,0.0,1.0);
	}
	else
	{
		normalize(v[0]);
		normalize(v[1]);
		normalize(v[2]);
	}
	// v[0]: horizontal   v[1]: vertical v[2]: screen normal
	setColumn(0, a_rotation, v[0]);
	setColumn(1, a_rotation, v[1]);
	setColumn(2, a_rotation, v[2]);
}

void ViewCommon::screenInfo(Real &a_screenScale,
		SpatialVector &a_screenLocation,
		const SpatialTransform &a_rotation,
		const SpatialVector &a_location)
{
	a_screenLocation = project(a_location, CameraI::e_current);
	a_screenLocation[2] = 0.0;

	SpatialVector horiz;
	getColumn(0,a_rotation,horiz);

	SpatialVector strutl = project(a_location + horiz, CameraI::e_current);
	strutl[2] = 0.0;

	a_screenScale = 1.0/magnitude(strutl - a_screenLocation);

}

//* static
Matrix<4,4,Real> ViewCommon::computeOrtho(Real a_zoom,
		const Vector<2,Real> &a_center,
		const Vector<2,Real> &a_size)
{
	FEASSERT(a_zoom>0.0f);
	FEASSERT(a_size[0]>0.0f);
	FEASSERT(a_size[1]>0.0f);

	//* NOTE get it pixel perfect

#if FALSE
	Real left = (-a_center[0]-0.375)/a_zoom;
	Real right = (-a_center[0]+a_size[0]-0.375)/a_zoom;
	Real bottom = (-a_center[1]-0.375)/a_zoom;
	Real top = (-a_center[1]+a_size[1]-0.375)/a_zoom;
#else
	Real left = (-a_center[0]*a_size[0]-0.375)/a_zoom;
	Real right = ((1.0-a_center[0])*a_size[0]-0.375)/a_zoom;
	Real bottom = (-a_center[1]*a_size[1]-0.375)/a_zoom;
	Real top = ((1.0-a_center[1])*a_size[1]-0.375)/a_zoom;
#endif

	//* TODO
	Real near_val = -1e3;
	Real far_val = 1e3;

	return ortho(left,right,bottom,top,near_val,far_val);
}

} // namespace
