import os
import re
import sys
import copy
import shutil
forge = sys.modules["forge"]

# NOTE: The Houdini version of boost may be different.
# The boost-regex appears to be sensitive and they don't provide the dso.
# Moving aside their include/boost seems to fix regex,
# but it leaves their other boost dso's cross-versioned with system includes.

# NOTE: Houdini ode includes also conflict
# Try moving aside their include/ode.

def setup(module):
	module.cppmap = {}
	test_include = {}
	test_cxx = ''
	test_cpp = {}
	test_link = {}
	test_variant = forge.houdini_sources[0][0]

	#NOTE separate obj path for each version

	# force all of FE to use Houdini's version of boost
#	forge.includemap['houdini'] = os.path.join(os.getenv('HT'),'include')

	deplibs = forge.corelibs + [
				"fexDrawDLLib",
				"fexMetaDLLib",
				"fexOperateDLLib",
				"fexSurfaceDLLib",
				"fexWindowLib"	]

	fe_houdini_oplibs = ':'.join(forge.houdini_oplibs_replace)
	if fe_houdini_oplibs == '':
		if os.path.exists('ext/grass'):
			fe_houdini_oplibs += ':fexGrassDL'
			sys.stdout.write(" grass")
		if os.path.exists('ext/ironworks'):
			fe_houdini_oplibs += ':fexIronWorksDL'
			sys.stdout.write(" ironworks")
		if os.path.exists('ext/vegetation'):
			fe_houdini_oplibs += ':fexVegetationDL'
			sys.stdout.write(" vegetation")
	else:
		fe_houdini_oplibs = ':' + fe_houdini_oplibs
	if len(forge.houdini_oplibs_append):
		fe_houdini_oplibs += ':' + ':'.join(forge.houdini_oplibs_append)

	manifest = 'ext/houdini/manifest.cc'
	with open(manifest, 'w') as outfile:
		outfile.write("")

	for houdini_source in forge.houdini_sources:
		if not os.path.exists(houdini_source[1]) or not os.path.exists(houdini_source[2]):
			forge.color_on(0, forge.BLUE)
			sys.stdout.write(" -" + houdini_source[0])
			forge.color_off()
			continue

		if houdini_source[0] == "":
			sys.stdout.write(" HT")
		else:
			sys.stdout.write(" " + houdini_source[0])

		variant = houdini_source[0]

		variantSplit = variant.split('.')
		if len(variantSplit) and variantSplit[0] != "":
			houdini_major = int(variantSplit[0])
			houdini_minor = int(variantSplit[1])
		else:
			houdini_major = 12
			houdini_minor = 0
			houdini_major_string = os.getenv('HOUDINI_MAJOR_RELEASE')
			if houdini_major_string:
				houdini_major = int(houdini_major_string)

		srcList = [	"houdini.pmh",
					"HoudiniBrush",
					"HoudiniDraw",
					"HoudiniSOP",
					"CacheOp",
					"PortalOp",
					"SurfaceAccessibleHoudini",
					"SurfaceAccessorHoudini",
					"SurfaceCurvesHoudini",
					"SurfacePointsHoudini",
					"SurfaceTrianglesHoudini",
					"assertHoudini",
					"houdiniDL" ]

		if houdini_major >= 15:
			srcList += [	"HoudiniPrimComponent",
							"HoudiniRenderComponent",
							"UnveilOp",
							"VeilOp" ]

		if variant != test_variant:
			variantPath = module.modPath + '/' + variant
			if os.path.lexists(variantPath) == 0:
				os.symlink('.', variantPath)

			srcListVariant = []
			for src in srcList:
				srcListVariant += [ variant + '/' + src ]
			srcList = srcListVariant

		dllname = "fexHoudiniDL" + variant

		dll = module.DLL( dllname, srcList )

		fe_houdini_libs = dllname + fe_houdini_oplibs

		alt_cxx = houdini_source[3]

		if alt_cxx != '':
			dll.cxx = alt_cxx

		# NOTE clearing all SSE flags may be excessively paranoid

		need_specialization = False
		clear_sse = False
		if forge.fe_os == "FE_LINUX":
			forge.evaluate_compiler(dll, False)
			# needed for gcc 4.2, not for gcc 4.4, but not sure otherwise
			if forge.compiler_version < 4.3:
				need_specialization = True
				clear_sse = True
			forge.evaluate_compiler(None, False)

		for src in srcList:
			relativePath = forge.RelativePath(forge.rootPath, module.modPath)
			modObjPath = os.path.join(forge.objPath, relativePath)
			objName = modObjPath + '/' + src
			if re.compile(r'(.*)\.pmh$').match(src):
				objName += '.gch'
			else:
				objName += forge.objsuf
			srcTarget = forge.FindTarget(objName)
			srcTarget.includemap = { 'houdini_src' : houdini_source[1] }

			if alt_cxx != '':
				srcTarget.cxx = alt_cxx

			srcTarget.cppmap = {}
			srcTarget.cppmap['houdini_src'] = '-DFE_HOUDINI_MAJOR=' + str(houdini_major)
			srcTarget.cppmap['houdini_src'] += ' -DFE_HOUDINI_MINOR=' + str(houdini_minor)
			srcTarget.cppmap['houdini_src'] += ' -DFE_HOUDINI_LIBS=' + fe_houdini_libs

			if houdini_major >= 15:
				srcTarget.cppmap['std'] = ' -std=c++11'

			if need_specialization:
				srcTarget.cppmap['houdini_src'] += ' -DNEED_SPECIALIZATION_STORAGE'

			if clear_sse:
				srcTarget.cppmap['sse'] = ''

			# NOTE later versions of openvdb require c++11

			if os.path.exists(houdini_source[1] + "openvdb/openvdb.h"):
				# houdini-supplied openvdb headers in 14+
				srcTarget.cppmap['houdini_src'] += ' -DFE_HOUDINI_USE_VDB=1'
			elif forge.openvdb_include != '' and forge.openvdb_include != 'auto' and houdini_major >= 15:
				srcTarget.cppmap['houdini_src'] += ' -DFE_HOUDINI_USE_VDB=1'

				srcTarget.includemap[ 'vdb' ] = os.path.join(forge.openvdb_include, '..')
				default_openexr = forge.openexr_source + '/include'
				if os.path.exists(default_openexr):
					srcTarget.includemap['openexr'] = default_openexr

			if variant == test_variant:
				test_include = copy.deepcopy(srcTarget.includemap)
				test_cpp = copy.deepcopy(srcTarget.cppmap)

				if alt_cxx != '':
					test_cxx = alt_cxx

		# this overrides previous -W flags in codegen
		module.cppmap = {}
		module.cppmap['houdini'] = ' -Wno-deprecated -Wall -W -Wno-parentheses -Wno-sign-compare -Wno-reorder -Wno-uninitialized -Wunused -Wno-unused-parameter'
		module.cppmap['houdini'] += ' -DFE_HOUDINI_DEFAULT_URL_ROOT=\\\"' + forge.houdini_url_root + '\\\"'
		module.cppmap['houdini'] += ' -DFE_HOUDINI_DEFAULT_ICON_ROOT=\\\"' + forge.houdini_icon_root + '\\\"'

		dll.cppmap = {}
		dll.linkmap = {}

		if clear_sse:
			dll.cppmap['sse'] = ''

		if forge.fe_os == "FE_LINUX":
			dll.linkmap["houdinilibs"] = " -Xlinker -rpath='$ORIGIN/../../'"
			dll.linkmap["houdinilibs"] += " -shared"
#			dll.linkmap["houdinilibs"] += " -Xlinker -rpath='" + houdini_source[2] + "'"
			dll.linkmap["houdinilibs"] += " -L" + houdini_source[2];
			# NOTE without explicit linkage, Mantra will whine
			dll.linkmap["houdinilibs"] += " -lHoudiniUI -lHoudiniOPZ -lHoudiniOP3 -lHoudiniOP2 -lHoudiniOP1 -lHoudiniSIM -lHoudiniGEO -lHoudiniPRM -lHoudiniUT -lHoudiniAPPS3"
		elif forge.fe_os == "FE_WIN32" or forge.fe_os == "FE_WIN64":
			dll.linkmap["houdinilibs"] = "TODO"

		if variant == test_variant:
			test_link = copy.deepcopy(dll.linkmap)

		forge.deps( [ dllname + "Lib" ], deplibs )

		# make sym links for houdini to look at
		# TODO should work on non-linux
		houdiniDso = 'lib' + dllname + '.so'
		houdiniPath = os.path.join(forge.libPath, 'houdini') + variant
		houdiniDsoPath = os.path.join(houdiniPath, 'dso')
		houdiniIconPath = os.path.join(houdiniPath, 'config', 'Icons')
		houdiniAppsPath = os.path.join(houdiniPath, 'config', 'Applications')
		houdiniUiPath = os.path.join(houdiniAppsPath, 'HoudiniBrush.ui')
		houdiniToolbarPath = os.path.join(houdiniPath, 'toolbar')
		if os.path.isdir(houdiniDsoPath) == 0:
			os.makedirs(houdiniDsoPath, 0755)
		if os.path.isdir(houdiniIconPath) == 0:
			os.makedirs(houdiniIconPath, 0755)
		if os.path.isdir(houdiniAppsPath) == 0:
			os.makedirs(houdiniAppsPath, 0755)
		if os.path.isdir(houdiniToolbarPath) == 0:
			os.makedirs(houdiniToolbarPath, 0755)

		source = os.path.join('..', '..', houdiniDso)

		dest = os.path.join(houdiniDsoPath, houdiniDso)
		if os.path.lexists(dest) == 0:
			os.symlink(source, dest)

		for icon in ['FE.svg', 'FE_alpha.svg', 'FE_beta.svg', 'FreeElectron.svg']:
			source = os.path.join('..','..','..','..','..','doc','image',icon)
			dest = os.path.join(houdiniIconPath, icon)
			if os.path.lexists(dest) == 0:
				os.symlink(source, dest)

		open(houdiniUiPath, 'a').close()

		shelf = os.path.join(houdiniToolbarPath, 'FreeElectron.shelf')
		shutil.copyfile('ext/houdini/FreeElectron.shelf', shelf)

		if houdini_major < 16:
			continue

		srcList = [	"houdiniEngine.pmh",
					"SurfaceAccessibleHapi",
					"SurfaceAccessorHapi",
					"SurfaceTrianglesHapi",
					"houdiniEngineDL" ]

		if variant != test_variant:
			srcListVariant = []
			for src in srcList:
				srcListVariant += [ variant + '/' + src ]
			srcList = srcListVariant

		dllname = "fexHoudiniEngineDL" + variant

		dll = module.DLL( dllname, srcList )

		for src in srcList:
			relativePath = forge.RelativePath(forge.rootPath, module.modPath)
			modObjPath = os.path.join(forge.objPath, relativePath)
			objName = modObjPath + '/' + src
			if re.compile(r'(.*)\.pmh$').match(src):
				objName += '.gch'
			else:
				objName += forge.objsuf
			srcTarget = forge.FindTarget(objName)
			srcTarget.includemap = { 'houdini_src' : houdini_source[1] }

			if alt_cxx != '':
				srcTarget.cxx = alt_cxx

			srcTarget.cppmap = {}
			srcTarget.cppmap['houdini_src'] = '-DFE_HOUDINI_DSO=' + houdini_source[2]
			srcTarget.cppmap['houdini_src'] += ' -DFE_HOUDINI_BUILD=' + houdiniDsoPath

		dll.cppmap = {}
		dll.linkmap = {}

		if clear_sse:
			dll.cppmap['sse'] = ''

		if forge.fe_os == "FE_LINUX":
			dll.linkmap["houdinilibs"] = " -Xlinker -rpath='$ORIGIN/../../'"
			dll.linkmap["houdinilibs"] += " -shared"
			dll.linkmap["houdinilibs"] += " -Xlinker -rpath='" + houdini_source[2] + "'"
			dll.linkmap["houdinilibs"] += " -L" + houdini_source[2];
			dll.linkmap["houdinilibs"] += " -lHAPIL"

		with open(manifest, 'a') as outfile:
			suffix = ""
			if variant != "":
				suffix = "." + variant
			for form in [ "hip", "hipnc" ]:
				outfile.write('\tspManifest->catalog<String>(\n'+
						'\t\t\t"SurfaceAccessibleI.SurfaceAccessibleHapi.fe.'+
						form + suffix + '")=\n'+
						'\t\t\t"fexHoudiniEngineDL' + variant + '";\n');

	test = module.Module('test')
	test.includemap = test_include
	test.cppmap = copy.deepcopy(module.cppmap)
	if 'houdini_src' in test_cpp:
		test.cppmap['houdini'] += " " + test_cpp['houdini_src']

	if test_cxx != '':
		test.cxx = test_cxx
		test.cppmap['sse'] = ''
	test.linkmap = copy.deepcopy(test_link)
	test_deplibs =	forge.corelibs + [
				"fexHoudiniDL" + test_variant + "Lib" ]

	tests = [ 'xHoudiniContext' ]
	for t in tests:
		test_exe = test.Exe(t)
		forge.deps([t + "Exe"], test_deplibs)

def auto(module):
	test_file = """
#include <SOP/SOP_API.h>

int main(void)
{
	return(0);
}
    \n"""

	houdini_path = os.getenv('HT')
	if houdini_path:
		forge.includemap['houdini'] = os.path.join(houdini_path,'include')

	result = forge.cctest(test_file)

	forge.includemap.pop('houdini', None)

	return result
