/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include <houdini/houdini.pmh>

#include "platform/dlCore.cc"

//* Houdini DSO
#include <UT/UT_DSOVersion.h>

using namespace fe;

extern "C"
{

FE_DL_EXPORT void HoudiniDSOInit(UT_DSOInfo& dsoinfo)
{
	//* force Houdini to load with RTLD_GLOBAL (needed for dynamic_cast<>)
	dsoinfo.loadGlobal=true;
}

FE_DL_EXPORT void ListDependencies(List<String*>& list)
{
//	printf("HoudiniDL ListDependencies\n");
	list.append(new String("fexMetaDL"));
}

FE_DL_EXPORT Library* CreateLibrary(sp<Master> a_spMaster)
{
	assertHoudini(a_spMaster->typeMaster());

//	printf("HoudiniDL CreateLibrary\n");
	Library *pLibrary = new Library();
	pLibrary->add<SurfaceCurvesHoudini>(
			"SurfaceI.SurfaceCurvesHoudini.fe");
	pLibrary->add<SurfacePointsHoudini>(
			"SurfaceI.SurfacePointsHoudini.fe");
	pLibrary->add<SurfaceTrianglesHoudini>(
			"SurfaceI.SurfaceTrianglesHoudini.fe");
	pLibrary->add<SurfaceAccessibleHoudini>(
			"SurfaceAccessibleI.SurfaceAccessibleHoudini.fe");
	pLibrary->add<SurfaceAccessorHoudini>(
			"SurfaceAccessorI.SurfaceAccessorHoudini.fe");
	pLibrary->add<HoudiniDraw>(
			"DrawI.HoudiniDraw.fe");
	pLibrary->add<HoudiniRamp>(
			"RampI.HoudiniRamp.fe");

	pLibrary->add<CacheOp>(
			"OperatorSurfaceI.CacheOp.fe");
	pLibrary->add<PortalOp>(
			"OperatorSurfaceI.PortalOp.fe.prototype");

#if FE_HOUDINI_15_PLUS
	pLibrary->add<UnveilOp>(
			"OperatorSurfaceI.UnveilOp.fe");
	pLibrary->add<VeilOp>(
			"OperatorSurfaceI.VeilOp.fe");
#endif

	return pLibrary;
}

FE_DL_EXPORT void InitializeLibrary(sp<Library> spLibrary)
{
}

}
