/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include <json/json.pmh>

#define FE_SAG_DEBUG	FALSE

namespace fe
{

/* string detail:

			[
				[
					"scope","public",
					"type","string",
					"name","bindDriverGroup",
					"options",{
					}
				],
				[
					"size",1,
					"storage","int32",
					"strings",[""
					],
					"indices",[
						"size",1,
						"storage","int32",
						"arrays",[[0]
						]
					]
				]
			],

	int detail:

			[
				[
					"scope","public",
					"type","numeric",
					"name","bindCount",
					"options",{
						"type":{
							"type":"string",
							"value":"nonarithmetic_integer"
						}
					}
				],
				[
					"size",1,
					"storage","int32",
					"defaults",[
						"size",1,
						"storage","fpreal64",
						"values",[0]
					],
					"values",[
						"size",1,
						"storage","int32",
						"arrays",[[8]
						]
					]
				]
			],
*/

//* static
Json::Value* SurfaceAccessorGeo::accessPointAttributes(
	sp<JsonValue> a_spJsonRoot,BWORD a_create)
{
	return &a_spJsonRoot->value()[13][3];
}

//* static
Json::Value* SurfaceAccessorGeo::accessVertexAttributes(
	sp<JsonValue> a_spJsonRoot,BWORD a_create)
{
	return &a_spJsonRoot->value()[13][1];
}

//* static
Json::Value* SurfaceAccessorGeo::accessPrimitiveAttributes(
	sp<JsonValue> a_spJsonRoot,BWORD a_create)
{
	if(a_spJsonRoot->value()[13].size()>4 &&
			a_spJsonRoot->value()[13][4]=="primitiveattributes")
	{
		return &a_spJsonRoot->value()[13][5];
	}
	if(a_spJsonRoot->value()[13].size()>6 &&
			a_spJsonRoot->value()[13][6]=="primitiveattributes")
	{
		return &a_spJsonRoot->value()[13][7];
	}
	return NULL;
}

//* static
Json::Value* SurfaceAccessorGeo::accessDetailAttributes(
	sp<JsonValue> a_spJsonRoot,BWORD a_create)
{
	if(a_spJsonRoot->value()[13].size()>4 &&
			a_spJsonRoot->value()[13][4]=="globalattributes")
	{
		return &a_spJsonRoot->value()[13][5];
	}
	if(a_spJsonRoot->value()[13].size()>6 &&
			a_spJsonRoot->value()[13][6]=="globalattributes")
	{
		return &a_spJsonRoot->value()[13][7];
	}
	if(!a_create)
	{
		return NULL;
	}
	const I32 index=(a_spJsonRoot->value()[12]=="primitiveattributes")? 14: 12;
	a_spJsonRoot->value()[index]="globalattributes";
	a_spJsonRoot->value()[index+1]=a_spJsonRoot->value()[index-1]; //* HACK
	return &a_spJsonRoot->value()[index+1][5];
}

//* static
Json::Value* SurfaceAccessorGeo::accessAttributes(
	sp<JsonValue> a_spJsonRoot,SurfaceAccessibleI::Element a_element,
	BWORD a_create)
{
	switch(a_element)
	{
		case SurfaceAccessibleI::e_point:
			return accessPointAttributes(a_spJsonRoot,a_create);
		case SurfaceAccessibleI::e_vertex:
			return accessVertexAttributes(a_spJsonRoot,a_create);
		case SurfaceAccessibleI::e_primitive:
			return accessPrimitiveAttributes(a_spJsonRoot,a_create);
		case SurfaceAccessibleI::e_detail:
			return accessDetailAttributes(a_spJsonRoot,a_create);
		default:
			;
	}

	return NULL;
}

I32 SurfaceAccessorGeo::elementCount(
	SurfaceAccessibleI::Element a_element) const
{
	switch(a_element)
	{
		case SurfaceAccessibleI::e_point:
			return m_spJsonRoot->value()[3].asInt();
		case SurfaceAccessibleI::e_vertex:
			return m_spJsonRoot->value()[5].asInt();
		case SurfaceAccessibleI::e_primitive:
			return m_spJsonRoot->value()[7].asInt();
		case SurfaceAccessibleI::e_detail:
			return 1;
		default:
			;
	}

	return 0;
}

void SurfaceAccessorGeo::addAttribute(String a_type,String a_typeValue,
	String a_storage,U32 a_components)
{
#if FE_SAG_DEBUG
	feLog("SurfaceAccessorGeo::addAttribute \"%s\"\n",
			m_attrName.c_str());
#endif

	Json::Value* pAttributes=accessAttributes(m_spJsonRoot,m_element,TRUE);
	if(!pAttributes)
	{
		feLog("SurfaceAccessorGeo::addInteger failed access %d\n",m_element);
		return;
	}

	const U32 attrIndex=(*pAttributes).size();

	Json::Value& rAttribute=(*pAttributes)[attrIndex];

	//* copy P as template
	Json::Value* pPointAttributes=
			accessAttributes(m_spJsonRoot,SurfaceAccessibleI::e_point,FALSE);
	rAttribute=(*pPointAttributes)[0];

	const U32 count=elementCount(m_element);

	rAttribute[0][3]=a_type.c_str();
	rAttribute[0][5]=m_attrName.c_str();
	if(!a_typeValue.empty())
	{
		rAttribute[0][7]["type"]["value"]=a_typeValue.c_str();
	}

	rAttribute[1][1]=a_components;			//* size
	rAttribute[1][3]=a_storage.c_str();		//* storage

	rAttribute[1][7][1]=a_components;		//* value size
	rAttribute[1][7][3]=a_storage.c_str();	//* value storage
	rAttribute[1][7][5].resize(0);
	rAttribute[1][7][5].resize(count);		//* values

	if(a_type=="string")
	{
		rAttribute[1][4]="strings";
		rAttribute[1][5].resize(0);
		rAttribute[1][5].resize(count);
		rAttribute[1][6]="indices";
		rAttribute[1][7][4]="arrays";

		for(U32 m=0;m<count;m++)
		{
			rAttribute[1][7][5][m][0]=m;
		}
	}
	else
	{
		rAttribute[1][5][1]=a_components;		//* default size
		rAttribute[1][5][5].resize(0);			//* default array
		rAttribute[1][5][5][0]=0;				//* default
	}

	m_pElementValues=&rAttribute[1][7][5];
	m_pStringValues=&rAttribute[1][5];
}

void SurfaceAccessorGeo::addString(void)
{
#if FE_SAG_DEBUG
	feLog("SurfaceAccessorGeo::addString \"%s\"\n",
			m_attrName.c_str());
#endif

	addAttribute("string","","int32",1);
}

void SurfaceAccessorGeo::addInteger(void)
{
#if FE_SAG_DEBUG
	feLog("SurfaceAccessorGeo::addInteger \"%s\"\n",
			m_attrName.c_str());
#endif

	addAttribute("numeric","nonarithmetic_integer","int32",1);
}

void SurfaceAccessorGeo::addReal(void)
{
#if FE_SAG_DEBUG
	feLog("SurfaceAccessorGeo::addReal \"%s\"\n",
			m_attrName.c_str());
#endif

	addAttribute("numeric","nonarithmetic_integer","fpreal32",1);
}

void SurfaceAccessorGeo::addSpatialVector(void)
{
#if FE_SAG_DEBUG
	feLog("SurfaceAccessorGeo::addSpatialVector \"%s\"\n",
			m_attrName.c_str());
#endif

	//* TODO typevalue: hpoint, normal, color, ...
	addAttribute("numeric","hpoint","fpreal32",4);
}

BWORD SurfaceAccessorGeo::bindInternal(SurfaceAccessibleI::Element a_element,
		const String& a_name)
{
#if FE_SAG_DEBUG
	feLog("SurfaceAccessorGeo::bindInternal \"%s\"\n",
			a_name.c_str());
#endif

	m_element=SurfaceAccessibleI::e_point;
	if(a_element<0 && a_element>5)
	{
		a_element=SurfaceAccessibleI::e_point;
	}

	m_element=a_element;
	m_attrName=a_name;
	m_pElementValues=NULL;

	String checkname=a_name;

	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		checkname="P";
		a_element=SurfaceAccessibleI::e_point;

		m_pVertexMap=&m_spJsonRoot->value()[11][1][1];
		m_pVertexValues=&m_spJsonRoot->value()[15][0][1];
	}

	Json::Value* pAttributes=accessAttributes(m_spJsonRoot,a_element,FALSE);
	if(!pAttributes)
	{
		return FALSE;
	}

	const U32 attrCount=(*pAttributes).size();
	for(U32 attrIndex=0;attrIndex<attrCount;attrIndex++)
	{
		Json::Value& rAttribute=(*pAttributes)[attrIndex];
		const String attrName=rAttribute[0][5].asString().c_str();

//		feLog("attr %d \"%s\"\n",attrIndex,attrName.c_str());

		if(attrName!=checkname)
		{
			continue;
		}

		m_pElementValues=&rAttribute[1][7][5];
		m_pStringValues=&rAttribute[1][5];

		return TRUE;
	}

	return FALSE;
}

} // namespace
