/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __lua_LuaIterate_h__
#define __lua_LuaIterate_h__

namespace fe
{

/** lua execution component.  Can also operate as a handler.

	@copydoc LuaIterate_info
	*/
class LuaIterate : public Initialize<LuaIterate>,
	virtual public Dispatch,
	virtual public HandlerI,
	virtual public LuaI,
	virtual public Config
{
	public:
		LuaIterate(void);
virtual	~LuaIterate(void);

		void	initialize(void);

		// AS LuaI
virtual	bool	preloadString(const String &text)			{ return false; }
virtual	bool	loadFile(const String &filename);
virtual	bool	loadString(const String &text);

virtual	void	set(const String &key, Record r_value);
virtual	void	set(const String &key, sp<RecordArray> ra_value);
virtual	void	set(const String &key, sp<RecordGroup> rg_value);
virtual	void	set(const String &key, const String &value);
virtual	void	set(const String &key, const Real &value)					{}
virtual	void	set(const String &key, Vector3 &value)						{}
virtual	void	set(const String &key, sp<Layout> l_value);
virtual	void	set(const String &key, sp<Component> spValue);
virtual	void	set(const String &key, sp<Scope> spValue);

virtual	Record	get(const String &key)	{ return Record(); }	//* TODO

virtual	bool	execute(void);
virtual	bool	execute(const String &a_fnName) { return false; };
virtual	void	flush(void);
virtual	void	alias(String a_aliasName,String a_trueName)					{}

		// AS HandlerI
virtual	void	handle(Record& r_sig);

		using Dispatch::call;

		// AS DispatchI
virtual	bool	call(const String &a_name, Array<Instance>& a_argv);

	private:
		bool				compile(void);
		lua_State			*m_pLuaState;
		String				m_chunk;
		sp<LuaContext>		m_spContext;
		String				m_name;
#if FE_COUNTED_TRACK
		Array<void*>		m_trackArray;
#endif
};

} /* namespace */


#endif /* __lua_LuaIterate_h__ */

