/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include <lua/lua.pmh>

#define FE_SAL_TICKER			TRUE

namespace fe
{

SurfaceAccessibleLua::SurfaceAccessibleLua(void)
{
//	feLog("SurfaceAccessibleLua::SurfaceAccessibleLua\n");
}

SurfaceAccessibleLua::~SurfaceAccessibleLua(void)
{
//	feLog("SurfaceAccessibleLua::~SurfaceAccessibleLua\n");
}

void SurfaceAccessibleLua::reset(void)
{
	SurfaceAccessibleRecord::reset();
}

BWORD SurfaceAccessibleLua::load(String a_filename,sp<Catalog> a_spSettings)
{
	sp<SurfaceAccessibleI> spDirectInput[4];
	const char* inputName[4]={ "input0", "input1", "input2", "input3" };

	Real frame=0.0;
	String options;
	String inputString;
	if(a_spSettings.isValid())
	{
		frame=a_spSettings->catalog<Real>("frame");
		options=a_spSettings->catalog<String>("options");
		inputString=a_spSettings->catalogOrDefault<String>("inputString","");

		for(I32 m=0;m<4;m++)
		{
			spDirectInput[m]=a_spSettings->catalogOrDefault< sp<Component> >(
					inputName[m],sp<Component>(NULL));
		}
	}

	if(m_spLuaI.isNull())
	{
		const String setupString=
				"function createPoint()\n"
				"point = pt:createRecord()\n"
				"output[\"points\"]:add(point)\n"
				"return point\n"
				"end\n"
				"function createPrimitive()\n"
				"primitive = prim:createRecord()\n"
				"output[\"primitives\"]:add(primitive)\n"
				"return primitive\n"
				"end\n"
				"function createVertex(primitive)\n"
				"point = createPoint()\n"
				"pointCount = output[\"points\"]:length()\n"
				"primitive[\"vertexMap\"]:add(pointCount-1)\n"
				"return point\n"
				"end\n\n"
				"function createOperator(implementation)\n"
				"terminalDispatch = create(\"*.TerminalDispatch\")\n"
				"terminalDispatch[\"_scope_\"] = scope\n"
				"terminalDispatch[\"_implementation_\"] = implementation\n"
				"return terminalDispatch\n"
				"end\n"
				"function replaceOutputWith(mesh)\n"
				"output[\"points\"] = mesh[\"points\"]\n"
				"output[\"vertices\"] = mesh[\"vertices\"]\n"
				"output[\"primitives\"] = mesh[\"primitives\"]\n"
				"output[\"detail\"] = mesh[\"detail\"]\n"
				"end\n\n";

		m_spLuaI=registry()->create("LuaI");

		m_spLuaI->preloadString(setupString);

		if(a_filename.empty() || !m_spLuaI->loadFile(a_filename))
		{
			if(inputString.empty() || !m_spLuaI->loadString(inputString))
			{
				feLog("SurfaceAccessibleLua::load"
						" failed to compile file or string\n");
				reset();
				return FALSE;
			}
		}

		//* TEMP should just pass along a unique scope to TerminalDispatch
//~		m_spScope=registry()->master()->catalog()->catalogComponent(
//~				"Scope","TerminalScope");

		m_spScope=registry()->create("Scope");

		m_spScope->setLocking(FALSE);

		bind(m_spScope);

		BWORD validInput=FALSE;

		sp<SurfaceAccessibleRecord> spRecordInput[4];
		for(I32 m=0;m<4;m++)
		{
			if(spDirectInput[m].isValid())
			{
				spRecordInput[m]=
						registry()->create("*.SurfaceAccessibleRecord");
				spRecordInput[m]->bind(m_spScope);
				spRecordInput[m]->copy(spDirectInput[m]);

				m_spLuaI->set(inputName[m],spRecordInput[m]->record());

				if(!validInput)
				{
					copy(spRecordInput[m]);
					validInput=TRUE;
				}
			}
		}

		if(!validInput)
		{
			reset();
		}

		m_spLuaI->set("scope",m_spScope);
		m_spLuaI->set("output",record());

		sp<Layout> spLayoutPoint=m_spScope->declare("pt");
		spLayoutPoint->populate("spc:at","vector3");
		m_spLuaI->set("pt",spLayoutPoint);

		sp<Layout> spLayoutPrimitive=m_spScope->declare("prim");
		spLayoutPrimitive->populate("surf:points","intarray");
		m_spLuaI->set("prim",spLayoutPrimitive);

		SurfaceModel surfaceModelRV;
		surfaceModelRV.bind(record());

		sp<RecordGroup> spRG=surfaceModelRV.surfacePointRG();
		RecordGroup::iterator it;
		if((it=spRG->begin()) != spRG->end())
		{
			m_spLuaI->set("pointArray",*it);
		}

		spRG=surfaceModelRV.surfaceVertexRG();
		if((it=spRG->begin()) != spRG->end())
		{
			m_spLuaI->set("vertexArray",*it);
		}

		spRG=surfaceModelRV.surfacePrimitiveRG();
		if((it=spRG->begin()) != spRG->end())
		{
			m_spLuaI->set("primitiveArray",*it);
		}

		spRG=surfaceModelRV.surfaceDetailRG();
		if((it=spRG->begin()) != spRG->end())
		{
			m_spLuaI->set("detailArray",*it);
		}

		m_spLuaI->alias("P","spc:at");
		m_spLuaI->alias("N","spc:up");
		m_spLuaI->alias("uv","surf:uvw");
		m_spLuaI->alias("vertexMap","surf:points");
		m_spLuaI->alias("points","surf:pointgroup");
		m_spLuaI->alias("vertices","surf:vertgroup");
		m_spLuaI->alias("primitives","surf:primgroup");
		m_spLuaI->alias("detail","surf:detgroup");
	}

	String option;
	while(!(option=options.parse()).empty())
	{
		const String property=option.parse("\"","=");
		const String value=option.parse("\"","=");

		if(value.match("^[-0-9.]+,[-0-9.]+,[-0-9.]+$"))
		{
			SpatialVector vector3(0.0,0.0,0.0);
			sscanf(value.c_str(),"%G,%G,%G",
					&vector3[0],&vector3[1],&vector3[2]);

//			feLog("\"%s\" vector %s\n",property.c_str(),c_print(vector3));
			m_spLuaI->set(property,vector3);
			continue;
		}
		if(value.match("^[-0-9.]+$"))
		{
			Real number=0.0;
			sscanf(value.c_str(),"%G",&number);

//			feLog("\"%s\" number %.6G\n",property.c_str(),number);
			m_spLuaI->set(property,number);
			continue;
		}

//		feLog("string \"%s\" \"%s\"\n",property.c_str(),value.c_str());
		m_spLuaI->set(property,value);
	}

	m_spLuaI->set("frame",frame);

#if FE_SAL_TICKER
	const U32 tickStart=systemTick();
#endif

	if(!m_spLuaI->execute())
	{
		feLog("SurfaceAccessibleLua::load Lua execution failed\n");
		return FALSE;
	}

#if FE_SAL_TICKER
	const U32 tickEnd=systemTick();
	const U32 tickDiff=SystemTicker::tickDifference(tickStart,tickEnd);
	const Real ms=1e-3*tickDiff*SystemTicker::microsecondsPerTick();
	feLog("SurfaceAccessibleLua::load executed %.6G ms\n",ms);
#endif

	Record result=m_spLuaI->get("output");

	m_spScope=result.layout()->scope();

	bind(result);

	feLog("SurfaceAccessibleLua::load points %d primitives %d\n",
			count(SurfaceAccessibleI::e_point),
			count(SurfaceAccessibleI::e_primitive));

	return TRUE;
}

} // namespace
