/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include <opencl/opencl.pmh>

#define FE_SAOC_CLONE_DEBUG		FALSE

namespace fe
{

sp<SurfaceAccessibleOpenCL::ClMem>
		SurfaceAccessibleOpenCL::Bridge::getClMem(void)
{
	safeLock();

#if FALSE
	feLog("SurfaceAccessibleOpenCL::Bridge::getClMem %d %p context %p\n",
			m_spClMem.isValid(),
			m_spClMem.isValid()? m_spClMem->m_clMem: nullptr,
			m_clContext);
#endif

	FEASSERT(m_spClMem.isNull() || m_spClMem->m_clMem!=nullptr);

	if(!m_gpuCurrent && m_spClMem.isValid())
	{
#if FALSE
		cl_command_queue clCommandQueue=(cl_command_queue)
				m_spMaster->catalog()->catalogOrDefault<void*>(
				"clCommandQueue",NULL);
		if(clCommandQueue)
		{
			cl_int clError=clEnqueueWriteBuffer(clCommandQueue,
					m_spClMem->m_clMem,CL_FALSE,0,m_elementCount*m_sizeof,
					m_bufferData,0,NULL,NULL);
			if(clError)
			{
				feLog("  clEnqueueWriteBuffer clError %d\n",
						clError);
			}
		}
		else
		{
			feLog("SurfaceAccessibleOpenCL::Bridge::getClMem"
					" no command queue\n");
			m_spClMem=NULL;
		}
#else
		//* just use clCreateBuffer below
		m_spClMem=NULL;
#endif
	}

	sp<ClMem> spClMem=m_spClMem;

	if(spClMem.isNull() && m_clContext)
	{
		spClMem=new ClMem();

		if(spClMem.isValid())
		{
			cl_int bufferSize=m_elementCount*m_sizeof;

			cl_int clError=0;
			spClMem->m_clMem=clCreateBuffer(m_clContext,
					CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR,
					bufferSize,m_bufferData,&clError);

#if FALSE
			feLog("SurfaceAccessibleOpenCL::Bridge::getClMem"
					" clCreateBuffer %p clError %d\n",
					spClMem->m_clMem,clError);
			feLog("  \"%s\" \"%s\" \"%s\" count %d sizeof %d\n",
					m_name.c_str(),m_type.c_str(),m_rate.c_str(),
					m_elementCount,m_sizeof);
#endif

			if(clError || spClMem->m_clMem==nullptr)
			{
				spClMem=NULL;
			}
		}

		m_spClMem=spClMem;
		m_gpuCurrent=m_spClMem.isValid();
	}

	safeUnlock();

	return spClMem;
}

void SurfaceAccessibleOpenCL::Bridge::refresh(void)
{
#if FALSE
	if(!m_cpuCurrent || !m_gpuCurrent)
	{
		feLog("SurfaceAccessibleOpenCL::Bridge::refresh %p"
				" \"%s\" cpu %d gpu %d\n",
				this,m_name.c_str(),m_cpuCurrent,m_gpuCurrent);
	}
#endif

	if(m_cpuCurrent)
	{
		return;
	}

	safeLock();

	if(!m_gpuCurrent)
	{
		feLog("SurfaceAccessibleOpenCL::Bridge::refresh"
				" neither side is current\n");
		return;
	}

	//* TODO other than vector3
//	if(m_type=="vector3")
	{
		cl_command_queue clCommandQueue=(cl_command_queue)
				m_spMaster->catalog()->catalogOrDefault<void*>(
				"clCommandQueue",NULL);

		if(clCommandQueue)
		{
			cl_mem clMem=(m_spClMem.isValid())? m_spClMem->m_clMem: NULL;

			if(clMem)
			{
				feLog("SurfaceAccessibleOpenCL::Bridge::refresh"
						" read back \"%s\"\n",m_name.c_str());

				cl_int clError;
				const size_t bufferSize=m_elementCount*m_sizeof;

#if FE_CODEGEN<=FE_DEBUG
				//* buffer size check
				size_t bytes=0;
				size_t returned=0;
				clError=clGetMemObjectInfo(clMem,CL_MEM_SIZE,
						sizeof(bytes),&bytes,&returned);
				if(clError)
				{
					feLog("  clGetMemObjectInfo clError %d\n",clError);
				}
				else if(bytes!=bufferSize)
				{
					feLog("SurfaceAccessibleOpenCL::Bridge::refresh"
							" cl_mem size %d expecting %d\n",
							bytes,bufferSize);
				}
#endif

				m_bufferData=reallocate(m_bufferData,bufferSize);

				clError=clEnqueueReadBuffer(clCommandQueue,
						clMem,CL_TRUE,0,bufferSize,
						m_bufferData,0,NULL,NULL);
				if(clError)
				{
					feLog("  clEnqueueReadBuffer clError %d\n",
							clError);

//					feLogBacktrace();
				}
			}
		}
	}

	m_cpuCurrent=TRUE;

	safeUnlock();
}

SurfaceAccessibleOpenCL::SurfaceAccessibleOpenCL(void):
	m_clContext(nullptr)
{
}

SurfaceAccessibleOpenCL::~SurfaceAccessibleOpenCL(void)
{
}

void SurfaceAccessibleOpenCL::reset(void)
{
	m_bridgeMap.clear();
}

I32 SurfaceAccessibleOpenCL::count(String a_nodeName,
	SurfaceAccessibleI::Element a_element) const
{
//	feLog("SurfaceAccessibleOpenCL::count element %d\n",a_element);

	String bridgeName;

	switch(a_element)
	{
		case e_point:
			bridgeName="pt:P";
			break;
		case e_primitive:
			bridgeName=":vertCount";
			break;
		default:
			;
	}

	if(bridgeName.empty())
	{
		return 0;
	}

	if(m_bridgeMap.find(bridgeName)==m_bridgeMap.end())
	{
//		feLog("SurfaceAccessibleOpenCL::count no bridge for \"%s\"\n",
//				bridgeName.c_str());
		return 0;
	}

	const sp<Bridge> spBridge=m_bridgeMap.at(bridgeName);
	if(spBridge.isNull())
	{
		feLog("SurfaceAccessibleOpenCL::count null bridge for \"%s\"\n",
				bridgeName.c_str());
		return 0;
	}

	return spBridge->m_elementCount;
}

Protectable* SurfaceAccessibleOpenCL::clone(Protectable* pInstance)
{
	SurfaceAccessibleOpenCL* pSurfaceAccessibleOpenCL= pInstance?
			dynamic_cast<SurfaceAccessibleOpenCL*>(pInstance):
			new SurfaceAccessibleOpenCL();

#if FE_SAOC_CLONE_DEBUG
	feLog("SurfaceAccessibleOpenCL::clone as %p from %p\n",
			pSurfaceAccessibleOpenCL,pInstance);
#endif

	updateContext();

	if(pSurfaceAccessibleOpenCL->library().isNull())
	{
		pSurfaceAccessibleOpenCL->setLibrary(library());
	}
	pSurfaceAccessibleOpenCL->m_clContext=m_clContext;

	for(std::map<String, sp<Bridge> >::iterator it=m_bridgeMap.begin();
			it!=m_bridgeMap.end();it++)
	{
		const String& bridgeName=it->first;
		const sp<Bridge>& rspSourceBridge=it->second;

#if FE_SAOC_CLONE_DEBUG
		feLog("  Bridge \"%s\"\n",bridgeName.c_str());
#endif

		if(rspSourceBridge.isValid())
		{
			sp<Bridge> spDestBridge=
					pSurfaceAccessibleOpenCL->lookupBridge(bridgeName,TRUE);
			spDestBridge->m_type=rspSourceBridge->m_type;
			spDestBridge->m_rate=rspSourceBridge->m_rate;
			spDestBridge->m_elementCount=rspSourceBridge->m_elementCount;
			spDestBridge->m_sizeof=rspSourceBridge->m_sizeof;

//			feLog("SurfaceAccessibleOpenCL::clone"
//					" \"%s\" src cpu %d gpu %d dest cpu %d gpu %d\n",
//					spDestBridge->m_name.c_str(),
//					rspSourceBridge->m_cpuCurrent,
//					rspSourceBridge->m_gpuCurrent,
//					spDestBridge->m_cpuCurrent,
//					spDestBridge->m_gpuCurrent);

			spDestBridge->m_cpuCurrent=FALSE;
			spDestBridge->m_gpuCurrent=FALSE;

			//* TODO shouldn't ever be zero
			const size_t entrySize=
					spDestBridge->m_sizeof? spDestBridge->m_sizeof: sizeof(I32);

			const cl_int bufferSize=spDestBridge->m_elementCount*entrySize;

			//* TODO check m_cpuCurrent and m_gpuCurrent

			if(rspSourceBridge->m_spClMem.isNull())
			{
				rspSourceBridge->getClMem();
			}

			if(rspSourceBridge->m_gpuCurrent &&
					rspSourceBridge->m_spClMem.isValid())
			{
				cl_mem clSourceMem=rspSourceBridge->m_spClMem->m_clMem;
				if(clSourceMem)
				{
					sp<ClMem> spNewClMem(new ClMem());

					if(spNewClMem.isValid())
					{

						cl_int clError=0;
						spNewClMem->m_clMem=clCreateBuffer(m_clContext,
								CL_MEM_READ_WRITE,
								bufferSize,nullptr,&clError);

#if FE_SAOC_CLONE_DEBUG
						feLog("SurfaceAccessibleOpenCL::clone"
								" clCreateBuffer %p clError %d\n",
								spNewClMem->m_clMem,clError);
#endif

						if(clError || spNewClMem->m_clMem==nullptr)
						{
							spNewClMem=NULL;
						}
						else
						{
							sp<Catalog> spMasterCatalog=
									registry()->master()->catalog();
							cl_command_queue clCommandQueue=(cl_command_queue)
									spMasterCatalog->catalogOrDefault<void*>(
									"clCommandQueue",NULL);

							clError=clEnqueueCopyBuffer(
									clCommandQueue,clSourceMem,
									spNewClMem->m_clMem,
									0,0,bufferSize,
									0,NULL,NULL);
#if FE_SAOC_CLONE_DEBUG
							feLog("  clEnqueueCopyBuffer %p to %p"
									" queue %p clError %d\n",
									clSourceMem,spNewClMem->m_clMem,
									clCommandQueue,clError);
#endif

							clError=clFinish(clCommandQueue);
#if FE_SAOC_CLONE_DEBUG
							if(clError)
							{
								feLog("  clFinish clError %d\n",clError);
							}
#endif
						}

						spDestBridge->m_gpuCurrent=TRUE;
					}

					spDestBridge->safeLock();
					spDestBridge->m_spClMemPrevious=spDestBridge->m_spClMem;
					spDestBridge->m_spClMem=spNewClMem;
					spDestBridge->safeUnlock();
				}
			}
			else if(!spDestBridge->m_gpuCurrent &&
					rspSourceBridge->m_cpuCurrent &&
					rspSourceBridge->m_bufferData)
			{
				spDestBridge->m_allocated=rspSourceBridge->m_allocated;
				const size_t allocSize=spDestBridge->m_allocated*entrySize;

#if FE_SAOC_CLONE_DEBUG
				feLog("    copy bufferData %d %d\n",bufferSize,allocSize);
#endif
				spDestBridge->m_bufferData=reallocate(nullptr,allocSize);

				memcpy(spDestBridge->m_bufferData,
						rspSourceBridge->m_bufferData,bufferSize);
				spDestBridge->m_cpuCurrent=TRUE;

				spDestBridge->safeLock();
				spDestBridge->m_spClMemPrevious=spDestBridge->m_spClMem;
				spDestBridge->m_spClMem=nullptr;
				spDestBridge->safeUnlock();
			}
		}
	}

#if FE_SAOC_CLONE_DEBUG
	feLog("SurfaceAccessibleOpenCL::clone result %p\n",
			pSurfaceAccessibleOpenCL);
#endif

	return pSurfaceAccessibleOpenCL;
}

void SurfaceAccessibleOpenCL::attributeSpecs(
	Array<SurfaceAccessibleI::Spec>& a_rSpecs,
	SurfaceAccessibleI::Element a_element) const
{
//	feLog("SurfaceAccessibleOpenCL::attributeSpecs element %d\n",a_element);

	a_rSpecs.clear();

	SurfaceAccessibleI::Spec spec;

	for(std::map<String, sp<Bridge> >::const_iterator it=m_bridgeMap.begin();
			it!=m_bridgeMap.end();it++)
	{
		const String& bridgeName=it->first;
		const sp<Bridge>& rspBridge=it->second;
		const String attrRate=rspBridge->m_rate.c_str();
		const String attrType=rspBridge->m_type.c_str();

//		feLog("spec Bridge \"%s\" type \"%s\" rate \"%s\"\n",bridgeName.c_str(),
//				attrType.c_str(),attrRate.c_str());

		if(attrType.empty() ||
				attrRate!=SurfaceAccessibleBase::elementLayout(a_element))
		{
			continue;
		}

		String text=bridgeName;
		text.parse("",":");

		const String attrName=text.parse("",":");
		if(attrName.empty())
		{
			continue;
		}

		spec.set(attrName,attrType);
		a_rSpecs.push_back(spec);
	}
}

sp<SurfaceAccessorI> SurfaceAccessibleOpenCL::accessor(
	String a_node,SurfaceAccessibleI::Element a_element,
	String a_name,SurfaceAccessibleI::Creation a_create)
{
	sp<SurfaceAccessorOpenCL> spAccessor=createAccessor();
	if(spAccessor.isValid())
	{
		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
		if(spAccessor->bind(a_element,a_name) ||
				a_create==SurfaceAccessibleI::e_createMissing)
		{
			return spAccessor;
		}
	}
	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleOpenCL::accessor(
	String a_node,SurfaceAccessibleI::Element a_element,
	SurfaceAccessibleI::Attribute a_attribute,
	SurfaceAccessibleI::Creation a_create)
{
	sp<SurfaceAccessorOpenCL> spAccessor=createAccessor();
	if(spAccessor.isValid())
	{
		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
		if(spAccessor->bind(a_element,a_attribute) ||
				a_create==SurfaceAccessibleI::e_createMissing)
		{
			return spAccessor;
		}
	}
	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleOpenCL::createAccessor(void)
{
	sp<SurfaceAccessorOpenCL> spAccessorOpenCL(new SurfaceAccessorOpenCL());

	return spAccessorOpenCL;
}

BWORD SurfaceAccessibleOpenCL::discard(
	SurfaceAccessibleI::Element a_element,String a_name)
{
	//* TODO
	return FALSE;
}

void SurfaceAccessibleOpenCL::updateContext(void)
{
	if(!m_clContext)
	{
		sp<Catalog> spMasterCatalog=registry()->master()->catalog();
		m_clContext=(cl_context)spMasterCatalog->catalogOrDefault<void*>(
				"clContext",nullptr);

		if(m_clContext)
		{
			for(std::map<String, sp<Bridge> >::iterator it=m_bridgeMap.begin();
					it!=m_bridgeMap.end();it++)
			{
				sp<Bridge>& rspBridge=it->second;
				if(rspBridge.isValid())
				{
					rspBridge->m_clContext=m_clContext;
				}
			}
		}
	}
}

sp<SurfaceAccessibleOpenCL::Bridge> SurfaceAccessibleOpenCL::lookupBridge(
	String a_name,BWORD a_create)
{
	updateContext();

	sp<Bridge> spBridge;

	if(a_create || m_bridgeMap.find(a_name)!=m_bridgeMap.end())
	{
		spBridge=m_bridgeMap[a_name];
		if(spBridge.isNull())
		{
			if(a_create)
			{
				spBridge=sp<Bridge>(new Bridge(m_clContext));
				spBridge->m_name=a_name;
				spBridge->m_spMaster=registry()->master();

				m_bridgeMap[a_name]=spBridge;
			}
			else
			{
				m_bridgeMap.erase(a_name);
			}
		}
	}

	return spBridge;
}

sp<SurfaceI> SurfaceAccessibleOpenCL::surface(String a_group)
{
	//* TODO select form
	//* TODO group
	sp<SurfaceI> spSurface;

	FEASSERT(registry().isValid());

	if(registry().isValid())
	{
		sp<SurfaceAccessorI> spSurfacePoints=
				accessor(e_point,e_position,e_refuseMissing);
		const I32 pointCount=spSurfacePoints.isValid()?
				spSurfacePoints->count(): 0;

		sp<SurfaceAccessorI> spSurfaceVertices=
				accessor(e_primitive,e_vertices,e_refuseMissing);
		const I32 primitiveCount=spSurfaceVertices.isValid()?
				spSurfaceVertices->count(): 0;

		const BWORD pointCloud=(pointCount && !primitiveCount);

		sp<SurfaceAccessorI> spSurfaceProperties=
				accessor(e_primitive,e_properties,e_refuseMissing);

		const BWORD openCurve=(spSurfaceProperties.isValid() &&
				spSurfaceProperties->count() &&
				spSurfaceProperties->integer(0,
				SurfaceAccessibleI::e_openCurve));

//		feLog("SurfaceAccessibleOpenCL::surface"
//				" \"%s\" pointCount %d primitiveCount %d"
//				" pointCloud %d openCurve %d\n",
//				a_group.c_str(),
//				pointCount,primitiveCount,pointCloud,openCurve);

		if(pointCloud || openCurve)
		{
//			feLog("SurfaceAccessibleOpenCL::surface"
//					" create SurfaceCurvesOpenCL\n");

			spSurface=registry()->create("*.SurfaceCurvesOpenCL");

			sp<SurfaceCurvesOpenCL> spSurfaceCurvesOpenCL=spSurface;
			if(spSurfaceCurvesOpenCL.isValid())
			{
				spSurfaceCurvesOpenCL->bind(
						sp<SurfaceAccessibleI>(this));
			}
		}
		else
		{
//			feLog("SurfaceAccessibleOpenCL::surface"
//					" create SurfaceTrianglesOpenCL\n");

			spSurface=registry()->create(
					"*.SurfaceTrianglesOpenCL");

			sp<SurfaceTrianglesOpenCL> spSurfaceTrianglesOpenCL=spSurface;
			if(spSurfaceTrianglesOpenCL.isValid())
			{
				spSurfaceTrianglesOpenCL->bind(
						sp<SurfaceAccessibleI>(this));
			}
		}
	}
	else
	{
		feLog("SurfaceAccessibleOpenCL::surface registry is NULL\n");
	}

	if(spSurface.isNull())
	{
		feLog("SurfaceAccessibleOpenCL::surface failed to access surface\n");
	}

	return spSurface;
}
} // namespace
