/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include <opencl/opencl.pmh>

#define	FE_SAOC_BIND_DEBUG			FALSE

#define	FE_SAOC_CHECK_LIMIT			FALSE
#define	FE_SAOC_CHECK_TYPECHANGE	FALSE

namespace fe
{

SurfaceAccessorOpenCL::SurfaceAccessorOpenCL(void)
{
	setName("SurfaceAccessorOpenCL");
}

SurfaceAccessorOpenCL::~SurfaceAccessorOpenCL(void)
{
}

void SurfaceAccessorOpenCL::setAttrType(String a_attrType)
{
#if FE_SAOC_CHECK_TYPECHANGE
	if(m_attrType==a_attrType)
	{
		return;
	}
	if(!m_attrType.empty())
	{
		feLog("SurfaceAccessorOpenCL::setAttrType"
				" changing type from \"%s\" to \"%s\"\n",
				m_attrType.c_str(),a_attrType.c_str());
	}
#else
	if(!m_attrType.empty())
	{
		return;
	}
#endif

	m_attrType=a_attrType;

	if(m_spBridge.isNull())
	{
		sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
				m_spSurfaceAccessibleI);
		if(spSurfaceAccessibleOpenCL.isNull())
		{
			return;
		}
		m_spBridge=spSurfaceAccessibleOpenCL->lookupBridge(m_bufferName,TRUE);
	}
	FEASSERT(m_spBridge.isValid());

	m_spBridge->m_type=a_attrType;
	m_spBridge->m_rate=
			SurfaceAccessibleBase::elementLayout(m_element);

	if(m_attrType=="integer")
	{
		m_spBridge->m_sizeof=sizeof(I32);
	}
	else if(m_attrType=="real")
	{
		m_spBridge->m_sizeof=sizeof(Real);
	}
	else if(m_attrType=="vector3")
	{
		m_spBridge->m_sizeof=4*sizeof(Real);
	}
}

U32 SurfaceAccessorOpenCL::count(void) const
{
	if(m_spBridge.isNull())
	{
		return m_spSurfaceAccessibleI->count(m_element);
	}

	m_spBridge->refresh();

	return m_spBridge->m_elementCount;
}

U32 SurfaceAccessorOpenCL::subCount(U32 a_index) const
{
	sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
			m_spSurfaceAccessibleI);
	if(spSurfaceAccessibleOpenCL.isNull())
	{
		return FALSE;
	}
	sp<SurfaceAccessibleOpenCL::Bridge> spVertCountBridge=
			spSurfaceAccessibleOpenCL->lookupBridge(":vertCount",FALSE);
	if(spVertCountBridge.isNull() ||
			I32(a_index)>=spVertCountBridge->m_elementCount)
	{
		return 0;
	}

//	feLog("SurfaceAccessorOpenCL::subCount(%d) current %d %d clMem %d %p\n",
//			a_index,
//			spVertCountBridge->m_cpuCurrent,
//			spVertCountBridge->m_gpuCurrent,
//			spVertCountBridge->m_spClMem.isValid(),
//			spVertCountBridge->m_spClMem.isValid()?
//			spVertCountBridge->m_spClMem->m_clMem: nullptr);

	spVertCountBridge->refresh();

//	feLog("SurfaceAccessorOpenCL::subCount buffer %p\n",
//			spVertCountBridge->m_bufferData);

	const I32* subCountArray=(I32*)(spVertCountBridge->m_bufferData);
	FEASSERT(subCountArray);

	return subCountArray[a_index];
}

void SurfaceAccessorOpenCL::ensureBufferSize(
	sp<SurfaceAccessibleOpenCL::Bridge> a_spBridge,I32 a_minSize,I32 a_sizeof)
{
	a_spBridge->refresh();

	a_spBridge->m_gpuCurrent=FALSE;

	if(!a_spBridge->m_sizeof)
	{
		a_spBridge->m_sizeof=a_sizeof;
	}

#if FALSE
	// size change
	if(a_spBridge->m_sizeof!=a_sizeof || a_sizeof<1)
	{
		feLog("SurfaceAccessorOpenCL::ensureBufferSize \"%s\" \"%s\" \"%s\""
				" %d vs %d allocated %d sizeof %d vs %d\n",
				a_spBridge->m_name.c_str(),
				m_attrType.c_str(),m_attrName.c_str(),
				a_minSize,a_spBridge->m_elementCount,a_spBridge->m_allocated,
				a_sizeof,a_spBridge->m_sizeof);
	}
#endif

	const I32 size=a_spBridge->m_elementCount;
	const I32 allocated=a_spBridge->m_allocated;
	if(a_minSize<=size && a_minSize<=allocated)
	{
		return;
	}

	if(a_minSize>allocated && a_sizeof>0)
	{
		//* tweak
		const I32 addCount=
				fe::maximum(I32(a_minSize-allocated),I32(0.2*allocated));

		const I32 oldBytes=a_spBridge->m_allocated*a_sizeof;
		a_spBridge->m_allocated+=addCount;
		const I32 newBytes=a_spBridge->m_allocated*a_sizeof;

#if FALSE
		feLog("SurfaceAccessorOpenCL::ensureBufferSize"
				" \"%s\" reallocate %d*%d for minCount %d\n",
				a_spBridge->m_name.c_str(),
				a_spBridge->m_allocated,a_sizeof,a_minSize);
#endif

		a_spBridge->m_bufferData=
				reallocate(a_spBridge->m_bufferData,newBytes);

		char* byteBuffer=(char*)(a_spBridge->m_bufferData);

		memset(byteBuffer+oldBytes,0,newBytes-oldBytes);
	}

	if(size<a_minSize)
	{
		a_spBridge->m_elementCount=a_minSize;
	}

//	FEASSERT(a_sizeof<1 || a_spBridge->m_elementCount<=a_spBridge->m_allocated);
}

void SurfaceAccessorOpenCL::set(U32 a_index,U32 subIndex,String a_string)
{
	setAttrType("string");

#if FE_SAOC_CHECK_LIMIT
	if(I32(a_index)>=count())
	{
		return;
	}
#endif

	//* TODO
}

String SurfaceAccessorOpenCL::string(U32 a_index,U32 a_subIndex)
{
	setAttrType("string");

	//* TODO

	return "";
}

void SurfaceAccessorOpenCL::set(U32 a_index,U32 a_subIndex,I32 a_integer)
{
	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
				m_spSurfaceAccessibleI);
		if(spSurfaceAccessibleOpenCL.isNull())
		{
			return;
		}

		sp<SurfaceAccessibleOpenCL::Bridge> spVertStartBridge=
				spSurfaceAccessibleOpenCL->lookupBridge(":vertStart",TRUE);
		if(spVertStartBridge.isNull())
		{
			return;
		}

		ensureBufferSize(spVertStartBridge,a_index+1,sizeof(I32));

		I32* subStartArray=(I32*)(spVertStartBridge->m_bufferData);

		sp<SurfaceAccessibleOpenCL::Bridge> spVertCountBridge=
				spSurfaceAccessibleOpenCL->lookupBridge(":vertCount",TRUE);
		if(spVertCountBridge.isNull())
		{
			return;
		}

		ensureBufferSize(spVertCountBridge,a_index+1,sizeof(I32));

		I32* subCountArray=(I32*)(spVertCountBridge->m_bufferData);

		sp<SurfaceAccessibleOpenCL::Bridge> spVertPointBridge=
				spSurfaceAccessibleOpenCL->lookupBridge(":vertPoint",TRUE);
		if(spVertPointBridge.isNull())
		{
			return;
		}

		//* TODO what is insertAt if substart and subcount are zero?

		I32& rVertCount=subCountArray[a_index];
		if(I32(a_subIndex)>=rVertCount)
		{
			const I32 insertAt=subStartArray[a_index]+rVertCount;
			const I32 insertCount=a_subIndex+1-rVertCount;

//			if(a_index<16)
//			feLog("    prim %d start %d sub %d/%d insert %d at %d for %d\n",
//					a_index,subStartArray[a_index],a_subIndex,rVertCount,
//					insertCount,insertAt,a_integer);

			ensureBufferSize(spVertPointBridge,
					spVertPointBridge->m_elementCount+insertCount,sizeof(I32));

			I32* vertexPointArray=(I32*)(spVertPointBridge->m_bufferData);

			const I32 insertTo=insertAt+insertCount;
			for(I32 vertexIndex=spVertPointBridge->m_elementCount-1;
					vertexIndex>=insertTo;vertexIndex--)
			{
				vertexPointArray[vertexIndex]=vertexPointArray[vertexIndex-1];
			}
			for(I32 vertexIndex=insertAt;vertexIndex<insertTo;vertexIndex++)
			{
				vertexPointArray[vertexIndex]= -1;
			}

			const I32 primitiveCount=spVertStartBridge->m_elementCount;
			for(I32 primitiveIndex=a_index+1;primitiveIndex<primitiveCount;
					primitiveIndex++)
			{
				subStartArray[primitiveIndex]+=insertCount;
			}

			rVertCount+=insertCount;
		}

		I32* vertexPointArray=(I32*)(spVertPointBridge->m_bufferData);

		const I32 vertexIndex=subStartArray[a_index]+a_subIndex;
		FEASSERT(vertexIndex<spVertPointBridge->m_elementCount);

		vertexPointArray[vertexIndex]=a_integer;

		return;
	}

	setAttrType("integer");

	FEASSERT(m_spBridge.isValid());

	ensureBufferSize(m_spBridge,a_index+1,sizeof(I32));

	I32* integerArray=(I32*)(m_spBridge->m_bufferData);

	integerArray[a_index]=a_integer;
}

I32 SurfaceAccessorOpenCL::integer(U32 a_index,U32 a_subIndex)
{
#if FALSE
	if(m_attribute==SurfaceAccessibleI::e_properties)
	{
		FEASSERT(m_element==
				SurfaceAccessibleI::e_primitiveGroup
				|| m_element==
				SurfaceAccessibleI::e_primitive);
		if(a_subIndex==SurfaceAccessibleI::e_openCurve)
		{
			//* TODO support non-curves
			return 1;
		}
		return 0;
	}
#endif

	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
				m_spSurfaceAccessibleI);
		if(spSurfaceAccessibleOpenCL.isNull())
		{
			return FALSE;
		}

		sp<SurfaceAccessibleOpenCL::Bridge> spVertStartBridge=
				spSurfaceAccessibleOpenCL->lookupBridge(
				":vertStart",FALSE);
		if(spVertStartBridge.isNull() ||
				I32(a_index)>=spVertStartBridge->m_elementCount)
		{
			return 0;
		}

		spVertStartBridge->refresh();

		const I32* subStartArray=(I32*)(spVertStartBridge->m_bufferData);

		const I32 vertexIndex=subStartArray[a_index]+a_subIndex;

		sp<SurfaceAccessibleOpenCL::Bridge> spVertPointBridge=
				spSurfaceAccessibleOpenCL->lookupBridge(":vertPoint",FALSE);
		if(spVertPointBridge.isNull() ||
				vertexIndex>=spVertPointBridge->m_elementCount)
		{
			return 0;
		}

		spVertPointBridge->refresh();

		const I32* vertexPointArray=(I32*)(spVertPointBridge->m_bufferData);

//		if(a_index<16)
//		feLog("prim %d sub %d lookup %d\n",
//				a_index,a_subIndex,vertexPointArray[vertexIndex]);

		return vertexPointArray[vertexIndex];
	}

	setAttrType("integer");

	FEASSERT(m_spBridge.isValid());

	m_spBridge->refresh();

	if(I32(a_index)>=m_spBridge->m_elementCount)
	{
		return 0;
	}

	const I32* integerArray=(I32*)(m_spBridge->m_bufferData);

	return integerArray[a_index];
}

//* coordinate with SurfaceAccessorBase::append(I32)
I32 SurfaceAccessorOpenCL::append(I32 a_integer)
{
//	feLog("SurfaceAccessorOpenCL::append(%d) '%s'\n",
//			a_integer,m_attrName.c_str());

	//* lot of points or primitives
	if(m_element==SurfaceAccessibleI::e_point ||
			(m_element==SurfaceAccessibleI::e_primitive &&
			m_attribute!=SurfaceAccessibleI::e_vertices))
	{
		if(a_integer<1)
		{
			return -1;
		}

		if(m_spBridge.isNull())
		{
			sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
					m_spSurfaceAccessibleI);
			if(spSurfaceAccessibleOpenCL.isNull())
			{
				return -1;
			}
			m_spBridge=spSurfaceAccessibleOpenCL->lookupBridge(
					m_bufferName,TRUE);
			m_spBridge->m_rate=
					SurfaceAccessibleBase::elementLayout(m_element);
		}
		FEASSERT(m_spBridge.isValid());

		const I32 first=count();

		switch(m_element)
		{
			case SurfaceAccessibleI::e_point:
				ensureBufferSize(m_spBridge,
						m_spBridge->m_elementCount+a_integer,
						m_spBridge->m_sizeof);
				break;
			case SurfaceAccessibleI::e_vertex:
				break;
			case SurfaceAccessibleI::e_primitive:
				ensureBufferSize(m_spBridge,
						m_spBridge->m_elementCount+a_integer,
						m_spBridge->m_sizeof);
				break;
			case SurfaceAccessibleI::e_detail:
				break;
			default:
				return -1;
		}

//		feLog("SurfaceAccessorOpenCL::append(%d) '%s' to %d\n",
//				a_integer,m_attrName.c_str(),first);

		return first;
	}

	if(m_element!=SurfaceAccessibleI::e_primitive ||
			m_attribute!=SurfaceAccessibleI::e_vertices)
	{
		append(U32(0),a_integer);
		return 0;
	}

	//* append a primitive with the specified number of vertices
	sp<SurfaceAccessorI> spPointAccessor=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);

	const I32 primitiveIndex=append();
	for(I32 vertIndex=0;vertIndex<a_integer;vertIndex++)
	{
		const I32 pointIndex=spPointAccessor->append();
		append(primitiveIndex,pointIndex);
	}

	return primitiveIndex;
}

I32 SurfaceAccessorOpenCL::append(void)
{
	//* point or primitive

	const U32 next=count();

//	feLog("SurfaceAccessorOpenCL::append() '%s' to %d\n",
//			m_attrName.c_str(),next);

	if(m_spBridge.isNull())
	{
		sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
				m_spSurfaceAccessibleI);
		if(spSurfaceAccessibleOpenCL.isNull())
		{
			return -1;
		}
		m_spBridge=spSurfaceAccessibleOpenCL->lookupBridge(
				m_bufferName,TRUE);
		m_spBridge->m_rate=
				SurfaceAccessibleBase::elementLayout(m_element);
	}
	FEASSERT(m_spBridge.isValid());

	switch(m_element)
	{
		case SurfaceAccessibleI::e_point:
			ensureBufferSize(m_spBridge,m_spBridge->m_elementCount+1,
					m_spBridge->m_sizeof);
			break;
		case SurfaceAccessibleI::e_vertex:
			break;
		case SurfaceAccessibleI::e_primitive:
			ensureBufferSize(m_spBridge,m_spBridge->m_elementCount+1,
					m_spBridge->m_sizeof);
			break;
		case SurfaceAccessibleI::e_detail:
			break;
		default:
			break;
	}

	return next;
}

I32 SurfaceAccessorOpenCL::append(SurfaceAccessibleI::Form a_form)
{
	//* TODO
	return -1;
}

void SurfaceAccessorOpenCL::append(U32 a_index,I32 a_integer)
{
	//* vertex of primitive
	const I32 vertIndex=subCount(a_index);
	set(a_index,vertIndex,a_integer);
}

//* coordinate with SurfaceAccessorBase::append(a_rPrimVerts)
void SurfaceAccessorOpenCL::append(
	Array< Array<I32> >& a_rPrimVerts)
{
	if(m_element!=SurfaceAccessibleI::e_primitive ||
			m_attribute!=SurfaceAccessibleI::e_vertices)
	{
		return;
	}

	const U32 appendPrimitiveCount=a_rPrimVerts.size();

//	feLog("SurfaceAccessorOpenCL::append '%s' primVerts %d\n",
//			m_attrName.c_str(),appendPrimitiveCount);

	sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
			m_spSurfaceAccessibleI);
	if(spSurfaceAccessibleOpenCL.isNull())
	{
		return;
	}

	sp<SurfaceAccessibleOpenCL::Bridge> spVertStartBridge=
			spSurfaceAccessibleOpenCL->lookupBridge(":vertStart",TRUE);
	if(spVertStartBridge.isNull())
	{
		return;
	}

	sp<SurfaceAccessibleOpenCL::Bridge> spVertCountBridge=
			spSurfaceAccessibleOpenCL->lookupBridge(":vertCount",TRUE);
	if(spVertCountBridge.isNull())
	{
		return;
	}

	sp<SurfaceAccessibleOpenCL::Bridge> spVertPointBridge=
			spSurfaceAccessibleOpenCL->lookupBridge(":vertPoint",TRUE);
	if(spVertPointBridge.isNull())
	{
		return;
	}

	spVertPointBridge->refresh();

	I32 priorVertexCount=spVertPointBridge->m_elementCount;

	I32 priorPrimitiveCount=spVertStartBridge->m_elementCount;
	const I32 toIndex=priorPrimitiveCount+appendPrimitiveCount-1;
	FEASSERT(toIndex>=0);

//	feLog("SurfaceAccessorOpenCL::append"
//			" prim from %d to %d priorVertexCount %d\n",
//			priorPrimitiveCount,toIndex,priorVertexCount);

	ensureBufferSize(spVertStartBridge,toIndex+1,sizeof(I32));
	ensureBufferSize(spVertCountBridge,toIndex+1,sizeof(I32));

	I32* subStartArray=(I32*)(spVertStartBridge->m_bufferData);
	for(I32 index=priorPrimitiveCount;index<=toIndex;index++)
	{
		subStartArray[index]=priorVertexCount;
	}

	I32* subCountArray=(I32*)(spVertCountBridge->m_bufferData);

	I32 appendVertexCount=0;
	for(U32 appendPrimitiveIndex=0;appendPrimitiveIndex<appendPrimitiveCount;
			appendPrimitiveIndex++)
	{
		const Array<I32>& rVertList=a_rPrimVerts[appendPrimitiveIndex];
		const U32 subCount=rVertList.size();
		appendVertexCount+=subCount;
	}

	const I32 newPrimitiveCount=priorPrimitiveCount+appendPrimitiveCount;
	const I32 newVertexCount=priorVertexCount+appendVertexCount;

	ensureBufferSize(spVertPointBridge,newVertexCount,sizeof(I32));
	I32* vertexPointArray=(I32*)(spVertPointBridge->m_bufferData);

	I32 primitiveIndex=priorPrimitiveCount;
	I32 vertexIndex=priorVertexCount;

	for(U32 appendPrimitiveIndex=0;appendPrimitiveIndex<appendPrimitiveCount;
			appendPrimitiveIndex++)
	{
		const Array<I32>& rVertList=a_rPrimVerts[appendPrimitiveIndex];
		const U32 subCount=rVertList.size();

		subStartArray[primitiveIndex]=vertexIndex;
		subCountArray[primitiveIndex]=subCount;

		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const I32 pointIndex=rVertList[subIndex];

//			if(primitiveIndex<8)
//			{
//				feLog("  primVert %d/%d %d/%d prim %d append %d\n",
//						appendPrimitiveIndex,appendPrimitiveCount,
//						subIndex,subCount,
//						primitiveIndex,pointIndex);
//			}

			FEASSERT(vertexIndex<newVertexCount);

			vertexPointArray[vertexIndex++]=pointIndex;
		}
		primitiveIndex++;
	}
}

void SurfaceAccessorOpenCL::set(U32 a_index,U32 a_subIndex,Real a_real)
{
	setAttrType("real");

	FEASSERT(m_spBridge.isValid());

	ensureBufferSize(m_spBridge,a_index+1,sizeof(Real));

	Real* realArray=(Real*)(m_spBridge->m_bufferData);

	realArray[a_index]=a_real;
}

Real SurfaceAccessorOpenCL::real(U32 a_index,U32 a_subIndex)
{
	setAttrType("real");

	FEASSERT(m_spBridge.isValid());

	m_spBridge->refresh();

	if(I32(a_index)<m_spBridge->m_elementCount)
	{
		const Real* realArray=(Real*)(m_spBridge->m_bufferData);
		return realArray[a_index];
	}

	return 0.0;
}

void SurfaceAccessorOpenCL::set(U32 a_index,U32 a_subIndex,
								const SpatialVector& a_spatialVector)
{
//	feLog("SurfaceAccessorOpenCL::set %d %d vector %s\n",
//			a_index,a_subIndex,c_print(a_spatialVector));

	setAttrType("vector3");

	FEASSERT(m_spBridge.isValid());

	if(m_attribute!=SurfaceAccessibleI::e_vertices)
	{
		ensureBufferSize(m_spBridge,a_index+1,4*sizeof(Real));

		Vector4* vectorArray=(Vector4*)(m_spBridge->m_bufferData);

		fe::set(vectorArray[a_index],
				a_spatialVector[0],a_spatialVector[1],a_spatialVector[2]);

		return;
	}

	const I32 pointIndex=integer(a_index,a_subIndex);

	sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
			m_spSurfaceAccessibleI);
	if(spSurfaceAccessibleOpenCL.isNull())
	{
		feLog("SurfaceAccessorOpenCL::set spatialVector"
				" not SurfaceAccessibleOpenCL\n");
		return;
	}

	sp<SurfaceAccessibleOpenCL::Bridge> spPositionBridge=
			spSurfaceAccessibleOpenCL->lookupBridge("pt:P",TRUE);
	if(spPositionBridge.isNull())
	{
		feLog("SurfaceAccessorOpenCL::set spatialVector no bridge data\n");
		return;
	}

	ensureBufferSize(spPositionBridge,pointIndex+1,4*sizeof(Real));

	Vector4* vectorArray=(Vector4*)(spPositionBridge->m_bufferData);
	fe::set(vectorArray[pointIndex],
			a_spatialVector[0],a_spatialVector[1],a_spatialVector[2]);
}

SpatialVector SurfaceAccessorOpenCL::spatialVector(
	U32 a_index,U32 a_subIndex)
{
//	setAttrType("vector3");

	FEASSERT(m_spBridge.isValid());

	m_spBridge->refresh();

#if FE_SAOC_CHECK_LIMIT
	if(a_index>=m_spBridge->m_elementCount)
	{
		return SpatialVector(0.0,0.0,0.0);
	}
#endif

	if(m_attribute!=SurfaceAccessibleI::e_vertices)
	{
		const Vector4& rVector4=
				((Vector4*)(m_spBridge->m_bufferData))[a_index];

//		if(a_index<8)
//		feLog("'%s' vector %d  %s\n",m_attrName.c_str(),
//				a_index,c_print(rVector4));

		return SpatialVector(rVector4[0],rVector4[1],rVector4[2]);
	}

	const I32 pointIndex=integer(a_index,a_subIndex);

	sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
			m_spSurfaceAccessibleI);
	if(spSurfaceAccessibleOpenCL.isNull())
	{
		feLog("SurfaceAccessorOpenCL::spatialVector"
				" not SurfaceAccessibleOpenCL\n");
		return SpatialVector(0.0,0.0,0.0);
	}

	sp<SurfaceAccessibleOpenCL::Bridge> spPositionBridge=
			spSurfaceAccessibleOpenCL->lookupBridge("pt:P",FALSE);
	if(spPositionBridge.isNull() ||
			pointIndex>=spPositionBridge->m_elementCount)
	{
		feLog("SurfaceAccessorOpenCL::spatialVector no bridge data\n");
		return SpatialVector(0.0,0.0,0.0);
	}

	spPositionBridge->refresh();

	const Vector4& rVector4=
			((Vector4*)(spPositionBridge->m_bufferData))[pointIndex];

//	if(a_index<8)
//	feLog("VERTEX '%s' vector %d %d pt %d  %s\n",m_attrName.c_str(),
//			a_index,a_subIndex,pointIndex,c_print(vectorArray[pointIndex]));

	return SpatialVector(rVector4[0],rVector4[1],rVector4[2]);
}

BWORD SurfaceAccessorOpenCL::bindInternal(
	SurfaceAccessibleI::Element a_element,const String& a_name)
{
	m_element=a_element;
	m_attrName=a_name;

	FEASSERT(a_element==SurfaceAccessibleI::e_point || a_name!="P");

#if FE_SAOC_BIND_DEBUG
	feLog("SurfaceAccessorOpenCL::bindInternal \"%s\" \"%s\"\n",
			SurfaceAccessibleBase::elementLayout(a_element).c_str(),
			a_name.c_str());
#endif

	//* TODO handle actual groups

	if(m_element==SurfaceAccessibleI::e_pointGroup)
	{
		if(a_name.empty())
		{
			m_element=SurfaceAccessibleI::e_point;
			m_attribute=SurfaceAccessibleI::e_position;
			m_attrName="P";
		}
		else
		{
			feLog("SurfaceAccessorOpenCL::bindInternal"
					" point groups not implemented (\"%s\")\n",
					a_name.c_str());
		}
	}

	if(m_element==SurfaceAccessibleI::e_primitiveGroup)
	{
		if(a_name.empty())
		{
			m_element=SurfaceAccessibleI::e_primitive;
			m_attribute=SurfaceAccessibleI::e_vertices;
			m_attrName="vertStart";
		}
		else
		{
			feLog("SurfaceAccessorOpenCL::bindInternal"
					" primitive groups not implemented (\"%s\")\n",
					a_name.c_str());
		}
	}

	sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
			m_spSurfaceAccessibleI);
	if(spSurfaceAccessibleOpenCL.isNull())
	{
#if FE_SAOC_BIND_DEBUG
		feLog("SurfaceAccessorOpenCL::bindInternal"
				" no SurfaceAccessibleOpenCL\n");
#endif
		return FALSE;
	}

	const String requestedRate=SurfaceAccessibleBase::elementLayout(m_element);

	m_bufferName=(m_attribute==SurfaceAccessibleI::e_vertices)?
			String(":vertStart"): requestedRate+":"+m_attrName;

	m_spBridge=spSurfaceAccessibleOpenCL->lookupBridge(m_bufferName,FALSE);
	if(m_spBridge.isNull())
	{
#if FE_SAOC_BIND_DEBUG
		feLog("SurfaceAccessorOpenCL::bindInternal no \"%s\"\n",
				m_bufferName.c_str());
#endif
		return FALSE;
	}

	if(m_attribute!=SurfaceAccessibleI::e_vertices)
	{
		m_spBridge->refresh();

		const String existingRate=m_spBridge->m_rate;
		if(existingRate!=requestedRate)
		{
#if FE_SAOC_BIND_DEBUG
			feLog("SurfaceAccessorOpenCL::bindInternal"
					" \"%s\" \"%s\" does not match existing rate \"%s\"\n",
					requestedRate.c_str(),m_attrName.c_str(),
					existingRate.c_str());
#endif
			return FALSE;
		}
	}

	return TRUE;
}

} // namespace
