/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include <opencl/opencl.pmh>

namespace fe
{

void SurfaceCurvesOpenCL::drawInternal(BWORD a_transformed,
	const SpatialTransform* a_pTransform,sp<DrawI> a_spDrawI,
	const fe::Color* a_pColor,sp<DrawBufferI> a_spDrawBuffer,
	sp<PartitionI> a_spPartition) const
{
	feLog("SurfaceCurvesOpenCL::drawInternal %p\n",
			dynamic_cast<const Component*>(this));

	sp<Catalog> spCatalog(a_spDrawBuffer);
	if(spCatalog.isValid())
	{
		sp<SurfaceAccessibleOpenCL> spSurfaceAccessibleOpenCL(
				m_spSurfaceAccessibleI);
		if(spSurfaceAccessibleOpenCL.isValid())
		{
			//* TODO other attributes
			for(I32 pass=0;pass<2;pass++)
			{
				const String bridgeName=pass? "pt:Cd": "pt:P";

				sp<SurfaceAccessibleOpenCL::Bridge> spBridge=
						spSurfaceAccessibleOpenCL->lookupBridge(
						bridgeName,FALSE);
				if(spBridge.isValid())
				{
					spBridge->safeLock();
					sp<SurfaceAccessibleOpenCL::ClMem> spClMem=
							spBridge->m_spClMemPrevious;
					spBridge->safeUnlock();

#if TRUE
					//* perhaps not dual-buffered
					if(spClMem.isNull())
					{
						spClMem=spBridge->getClMem();
					}
#endif

					if(spBridge->m_elementCount!=I32(m_vertices))
					{
//						feLog("SurfaceCurvesOpenCL::drawInternal"
//								" bridge has %d elements vs %d verts"
//								" (not simple curves)\n",
//								spBridge->m_elementCount,m_vertices);
						spClMem=NULL;
					}

					sp<SurfaceAccessibleOpenCL::ClMem> spClMemNext=
							spBridge->getClMem();
					feLog("SurfaceCurvesOpenCL::drawInternal"
							" \"%s\" mem %p %p\n",
							bridgeName.c_str(),
							spClMem.isValid()? spClMem->m_clMem: nullptr,
							spClMemNext.isValid()? spClMemNext->m_clMem: nullptr);

					if(spClMem.isValid())
					{
						spCatalog->catalog< sp<Counted> >("cl:"+bridgeName)=
								spClMem;
					}
				}
			}
		}
	}

	SurfaceCurvesAccessible::drawInternal(a_transformed,a_pTransform,
			a_spDrawI,a_pColor,a_spDrawBuffer,a_spPartition);
}

} // namespace
