import os
import stat
import sys
import re
forge = sys.modules["forge"]

def setup(module):

	version = ""
	for line in open(forge.opencl_include + "/CL/cl.h").readlines():
		if re.match("#define CL_VERSION_.*", line):
			version = line.split()[1][11:].rstrip().replace("_",".")

	if version != "":
		sys.stdout.write(" " + version)

	forge.cppmap['opencl'] = '-DFE_USE_OPENCL'
	forge.cppmap['opencl'] += ' -DFE_GL_OPENCL'

	srcList = [ "opencl.pmh",
				"CacheCLOp",
				"ContractCLOp",
				"ClumpCLOp",
				"CurlCLOp",
				"FlatnessCLOp",
				"KinkCLOp",
				"OpenCLContext",
				"OpenCLOp",
				"SurfaceAccessibleOpenCL",
				"SurfaceAccessorOpenCL",
				"SurfaceCurvesOpenCL",
				"SurfaceTrianglesOpenCL",
				"TransformCLOp",
				"openclDL" ]

	relativePath = forge.RelativePath(forge.rootPath,module.modPath)
	modObjPath = os.path.join(forge.objPath, relativePath)
	clcPath = modObjPath + "/clc"
	if os.path.isdir(clcPath) == 0:
		os.makedirs(clcPath, 0755)

	sourceFiles = os.listdir(module.modPath)
	for sourceFile in sourceFiles:
		if re.compile(r'(.*)\.cl$').match(sourceFile):
			clFile = module.modPath + "/" + sourceFile
			clcFile = clcPath + "/" + sourceFile + ".c"

			clModification = os.stat(clFile)[stat.ST_MTIME]
			if os.path.lexists(clcFile):
				clcModification = os.stat(clcFile)[stat.ST_MTIME]
			else:
				clcModification = 0

			if clModification > clcModification:
				command = module.modPath + "/preprocess.py " + clFile + " " + clcFile
				os.system(command)

	dll = module.DLL( "fexOpenCLDL", srcList )

	module.includemap = {}
	module.includemap['opencl'] = forge.opencl_include
	module.includemap['clc'] = clcPath

	module.cppmap = { 'std' : "-std=c++11" }

	dll.linkmap = {}
	dll.linkmap['opencl'] = "-Xlinker -rpath='" + forge.opencl_lib + "'"
	dll.linkmap['opencl'] += ' -L' + forge.opencl_lib + ' -lOpenCL'

	deplibs = forge.corelibs + [
				"feSignalLib",
				"fexMathLib",
				"fexGeometryDLLib",
				"fexOperatorDLLib",
				"fexDataToolDLLib",
				"fexWindowLib"	]

	forge.deps( ["fexOpenCLDLLib"], deplibs )

def auto(module):
	test_file = """
#include <CL/cl.h>

int main(void)
{
	return(0);
}
    \n"""

	# TODO also configurable (as above)
	forge.includemap['opencl'] = forge.opencl_include

	result = forge.cctest(test_file)

	forge.includemap.pop('opencl', None)

	return result
