/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_SurfaceAccessibleRecord_h__
#define __surface_SurfaceAccessibleRecord_h__

namespace fe
{

/**************************************************************************//**
    @brief Record Surface Binding

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessibleRecord:
	public ObjectSafeShared<SurfaceAccessibleRecord>,
	virtual public SurfaceAccessibleBase
{
	public:
								SurfaceAccessibleRecord(void);
virtual							~SurfaceAccessibleRecord(void);

								//* As Protectable
virtual	Protectable*			clone(Protectable* pInstance=NULL);

								//* as SurfaceAccessibleI
virtual	void					bind(Instance a_instance)
								{
									Record record=a_instance.cast<Record>();
									if(record.isValid())
									{
										sp<Scope> spScope=
												record.layout()->scope();
										bind(spScope);
									}
									bind(record);
								}
virtual	BWORD					isBound(void)
								{	return m_weakRecord.isValid(); }

								using SurfaceAccessibleBase::count;

virtual	I32						count(String a_nodeName,
										SurfaceAccessibleI::Element
										a_element) const;

virtual	void					bind(sp<Scope> a_spScope)
								{	m_spScope=a_spScope;
									reset(); }
virtual	BWORD					load(String a_filename)
								{	return load(a_filename,sp<Catalog>(NULL)); }
virtual	BWORD					load(String a_filename,
										sp<Catalog> a_spSettings);
virtual	BWORD					save(String a_filename)
								{	return save(a_filename,sp<Catalog>(NULL)); }
virtual	BWORD					save(String a_filename,
										sp<Catalog> a_spSettings);

virtual	void					bind(sp<SurfaceAccessibleI::ThreadingState>
										a_spThreadingState)
								{	m_spThreadingState=a_spThreadingState; }

virtual	void					groupNames(Array<String>& a_rNameArray,
										SurfaceAccessibleI::Element
										a_element) const;

								using SurfaceAccessibleBase::attributeSpecs;

virtual	void					attributeSpecs(
										Array<SurfaceAccessibleI::Spec>&
										a_rSpecs,
										SurfaceAccessibleI::Element
										a_element) const;

virtual void					require(SurfaceAccessibleI::Element a_element,
										String a_name,String a_typeName);

								using SurfaceAccessibleBase::accessor;

virtual sp<SurfaceAccessorI>	accessor(String a_node,Element a_element,
										String a_name,Creation a_create);

virtual sp<SurfaceAccessorI>	accessor(String a_node,Element a_element,
										Attribute a_attribute,
										Creation a_create);

virtual	sp<SurfaceI>			surface(void)
								{	return surface(String()); }
virtual	sp<SurfaceI>			surface(String a_group);
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex)
								{	return subSurface(a_subIndex,String()); }
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex,String a_group);

virtual	void					copy(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible);
virtual	void					copy(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible,
										sp<FilterI> a_spFilter);

								//* as SurfaceAccessibleRecord
		void					bind(Record& a_rRecord);
		Record					record(void)
								{	return m_weakRecord; }

		void					bindLike(sp<SurfaceAccessibleRecord>& a_rspOther)
								{
									if(a_rspOther.isValid())
									{
										m_spThreadingState=
												a_rspOther->m_spThreadingState;
										m_paging=a_rspOther->m_paging;
										m_spScope=a_rspOther->m_spScope;
										m_spRecordGroup=
												a_rspOther->m_spRecordGroup;
										m_weakRecord=a_rspOther->m_weakRecord;
									}
								}
virtual	sp<Component>			payload(void)
								{	return m_spForeignAccessible; }

	protected:
virtual	void					reset(void);

		sp<Scope>				m_spScope;
		sp<RecordGroup>			m_spRecordGroup;
		WeakRecord				m_weakRecord;
		sp<SurfaceAccessibleI>	m_spForeignAccessible;
};

} // namespace

#endif /* __surface_SurfaceAccessibleRecord_h__ */
