/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include <usd/usd.pmh>

#define FE_SAU_DEBUG	FALSE

namespace fe
{

void SurfaceAccessibleUsd::reset(void)
{
	m_usdStage=NULL;
}

sp<SurfaceAccessorI> SurfaceAccessibleUsd::accessor(
	String a_node,Element a_element,String a_name,Creation a_create)
{
#if FE_SAU_DEBUG
	if(a_name!="instance")
	{
		feLog("SurfaceAccessibleUsd::accessor"
				" node \"%s\" elem '%s' attr \"%s\" create %d\n",
				a_node.c_str(),
				SurfaceAccessibleBase::elementLayout(a_element).c_str(),
				a_name.c_str(),
				a_create==SurfaceAccessibleI::e_createMissing);
	}
#endif

	if(a_node.empty())
	{
		sp<SurfaceAccessorUsdGraph> spAccessor(new SurfaceAccessorUsdGraph);

		if(spAccessor.isValid())
		{
			spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
			if(spAccessor->bind(a_element,a_name) ||
					a_create==SurfaceAccessibleI::e_createMissing)
			{
				return spAccessor;
			}
		}
	}
	else
	{
		sp<SurfaceAccessorUsd> spAccessor(new SurfaceAccessorUsd);

		if(spAccessor.isValid())
		{
			UsdPrim usdPrim;

			sp<PrimNode> spPrimNode=m_nodeMap[a_node];
			if(spPrimNode.isValid())
			{
				usdPrim=spPrimNode->usdPrim();
			}

			if(!usdPrim)
			{
				return sp<SurfaceAccessorI>(NULL);
			}

			spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
			spAccessor->setUsdPrim(usdPrim);
			spAccessor->setFrame(m_frame);
			if(spAccessor->bind(a_element,a_name) ||
					a_create==SurfaceAccessibleI::e_createMissing)
			{
				return spAccessor;
			}
		}
	}

#if FE_SAU_DEBUG
	if(a_name!="instance")
	{
		feLog("SurfaceAccessibleUsd::accessor fail\n");
	}
#endif

	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleUsd::accessor(
	String a_node,Element a_element,Attribute a_attribute,Creation a_create)
{
#if FE_SAU_DEBUG
	feLog("SurfaceAccessibleUsd::accessor"
			" node \"%s\" elem '%s' attr '%s' create %d\n",
			a_node.c_str(),
			SurfaceAccessibleBase::elementLayout(a_element).c_str(),
			SurfaceAccessibleBase::attributeString(a_attribute).c_str(),
			a_create==SurfaceAccessibleI::e_createMissing);
#endif

	if(a_node.empty())
	{
		sp<SurfaceAccessorUsdGraph> spAccessor(new SurfaceAccessorUsdGraph);

		if(spAccessor.isValid())
		{
			spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
			if(spAccessor->bind(a_element,a_attribute) ||
					a_create==SurfaceAccessibleI::e_createMissing)
			{
				return spAccessor;
			}
		}
	}
	else
	{
		sp<SurfaceAccessorUsd> spAccessor(new SurfaceAccessorUsd);

		if(spAccessor.isValid())
		{
			UsdPrim usdPrim;

			sp<PrimNode> spPrimNode=m_nodeMap[a_node];
			if(spPrimNode.isValid())
			{
				usdPrim=spPrimNode->usdPrim();
			}

			if(!usdPrim)
			{
				return sp<SurfaceAccessorI>(NULL);
			}

			spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
			spAccessor->setUsdPrim(usdPrim);
			spAccessor->setFrame(m_frame);
			if(spAccessor->bind(a_element,a_attribute) ||
					a_create==SurfaceAccessibleI::e_createMissing)
			{
				return spAccessor;
			}
		}
	}

#if FE_SAU_DEBUG
	feLog("SurfaceAccessibleUsd::accessor fail\n");
#endif

	return sp<SurfaceAccessorI>(NULL);
}

BWORD SurfaceAccessibleUsd::load(String a_filename,sp<Catalog> a_spSettings)
{
#if FE_SAU_DEBUG
	feLog("SurfaceAccessibleUsd::load \"%s\"\n",a_filename.c_str());
#endif

	String options;

	if(a_spSettings.isValid())
	{
		m_frame=a_spSettings->catalog<Real>("frame");
		options=a_spSettings->catalog<String>("options");
//		spFilter=a_spSettings->catalog< sp<Component> >("filter");
	}
	else
	{
		m_frame=0.0;
		options="";
	}

	std::map<String,String> optionMap;

	String buffer=options;
	String option;
	while(!(option=buffer.parse()).empty())
	{
		const String property=option.parse("\"","=");
		const String value=option.parse("\"","=");

#if FE_SAXG_DEBUG
		feLog("SurfaceAccessibleXGen::load parse \"%s\" \"%s\"\n",
				property.c_str(),value.c_str());
#endif

		optionMap[property]=value;
	}

	m_nodeArray.clear();
	m_nodeMap.clear();

    outlineClear();

	m_usdStage=UsdStage::Open(a_filename.c_str());
	if(!m_usdStage)
	{
		feLog("SurfaceAccessibleUsd::load failed\n");
		return FALSE;
	}

	UsdPrim rootPrim=m_usdStage->GetPseudoRoot();

#if FE_SAU_DEBUG
	const TfToken& rootName=rootPrim.GetName();
	const TfToken& rootType=rootPrim.GetTypeName();
	feLog("root \"%s\" type \"%s\"\n",rootName.GetText(),rootType.GetText());

	UsdPrim defPrim=m_usdStage->GetDefaultPrim();
	if(defPrim.IsValid())
	{
		const TfToken& defName=defPrim.GetName();
		const TfToken& defType=defPrim.GetTypeName();
		feLog("def \"%s\" type \"%s\"\n",defName.GetText(),defType.GetText());
	}
#endif

//	for(UsdTreeIterator it(rootPrim); it ; ++it)

	UsdPrim::SubtreeRange subtreeRange=rootPrim.GetAllDescendants();
	for(UsdPrim::SubtreeRange::iterator it=subtreeRange.begin();
			it!=subtreeRange.end();it++)
	{
		const UsdPrim& usdPrim= *it;

		if(!usdPrim.IsValid())
		{
			continue;
		}

		const UsdPrim primParent=usdPrim.GetParent();

		const String primType=usdPrim.GetTypeName().GetText();
		const String primName=usdPrim.GetName().GetText();

#if FE_SAU_DEBUG
		feLog("\nprim \"%s\" type \"%s\"\n",
				primName.c_str(),primType.c_str());
#endif

		//* TODO get full path for key
		String primPath="/"+primName;
		String parentPath;
		UsdPrim ancestor=primParent;
		while(ancestor)
		{
			const String ancestorName=ancestor.GetName().GetText();
			if(ancestorName=="/")
			{
				break;
			}

			primPath="/"+ancestorName+primPath;
			parentPath="/"+ancestorName+parentPath;

			ancestor=ancestor.GetParent();
		}

		sp<PrimNode> spPrimNode(new PrimNode());
		m_nodeArray.push_back(spPrimNode);
		m_nodeMap[primPath]=spPrimNode;

		spPrimNode->setUsdPrim(usdPrim);
		spPrimNode->setUsdType(primType);
		spPrimNode->setUsdPath(primPath);
		spPrimNode->setUsdParent(parentPath);

		const BWORD isMesh=(primType=="Mesh");
		const BWORD outline=isMesh;

		if(outline)
		{
			outlineAppend(primPath);
			outlineAppend("  ("+primType+")");
		}

		Array<SurfaceAccessibleI::Spec>& rPointSpecs=
				spPrimNode->pointSpecs();
		Array<SurfaceAccessibleI::Spec>& rVertexSpecs=
				spPrimNode->vertexSpecs();
		Array<SurfaceAccessibleI::Spec>& rPrimitiveSpecs=
				spPrimNode->primitiveSpecs();
		Array<SurfaceAccessibleI::Spec>& rDetailSpecs=
				spPrimNode->detailSpecs();

#if FE_SAU_DEBUG
		UsdMetadataValueMap metaMap=usdPrim.GetAllMetadata();
		for(UsdMetadataValueMap::const_iterator it=metaMap.begin();
				it!=metaMap.end(); it++)
		{
			const TfToken& rToken=it->first;
//			const VtValue& rValue=it->second;
			feLog("META \"%s\"\n",rToken.GetText());
		}
#endif

		I32 pointCount=0;
		I32 vertexCount=0;
		I32 primitiveCount=0;

		std::vector<UsdProperty> propArray=usdPrim.GetProperties();
		const I32 propCount=propArray.size();
		for(I32 propIndex=0;propIndex<propCount;propIndex++)
		{
			UsdProperty& rProperty=propArray[propIndex];
			const String propName=rProperty.GetName().GetText();
			String line="  "+propName;

#if FE_SAU_DEBUG
			feLog("  prop %d/%d \"%s\" ",propIndex,propCount,propName.c_str());
#endif

			const BWORD isPrimVar=!strncmp(propName.c_str(),"primvars:",9);

			if(UsdAttribute attr=rProperty.As<UsdAttribute>())
			{
				const String roleName=attr.GetRoleName().GetText();
				const String typeName=
						attr.GetTypeName().GetType().GetTypeName().c_str();

				if(!roleName.empty())
				{
					line.cat(" (%s)",roleName.c_str());
				}
				if(!typeName.empty())
				{
					line.cat(" %s",typeName.c_str());
				}

				//* TODO rate?

#if FE_SAU_DEBUG
				feLog("type \"%s\" role \"%s\"\n",
						typeName.c_str(),roleName.c_str());
#endif

#if FE_SAU_DEBUG
				feLog("type \"%s\" role \"%s\"\n",
						typeName.c_str(),roleName.c_str());
#endif

				std::vector<double> timeArray;
				attr.GetTimeSamples(&timeArray);

				const I32 timeCount=timeArray.size();
#if FE_SAU_DEBUG
				if(timeCount>2)
				{
					feLog("      time %.6G..%.6G\n",
							timeArray[0],timeArray[timeCount-1]);
				}
				else if(timeCount)
				{
					feLog("      time %.6G\n",timeArray[0]);
				}
#endif

				const UsdTimeCode timeCode=timeCount?
						UsdTimeCode(timeArray[0]): UsdTimeCode::Default();

				SurfaceAccessibleI::Spec spec;
				I32 valueCount=0;

				if(typeName=="TfToken")
				{
					TfToken value;
					if(attr.Get(&value,timeCode))
					{
#if FE_SAU_DEBUG
						feLog("      value \"%s\"\n",value.GetText());
#endif

						spec.set(propName,"string");
					}
				}
				else if(typeName=="bool")
				{
					bool value;
					if(attr.Get(&value,timeCode))
					{
#if FE_SAU_DEBUG
						feLog("      value %d\n",value);
#endif

						spec.set(propName,"integer");
					}
				}
				else if(typeName=="double")
				{
					double value;
					if(attr.Get(&value,timeCode))
					{
#if FE_SAU_DEBUG
						feLog("      value %.6G\n",value);
#endif

						spec.set(propName,"real");
					}
				}
				else if(typeName=="VtArray<TfToken>")
				{
					VtArray<TfToken> array;
					if(attr.Get(&array,timeCode))
					{
						valueCount=array.size();
#if FE_SAU_DEBUG
						feLog("      size %d \n",valueCount);
						for(I32 index=0;index<valueCount && index<16;index++)
						{
							const TfToken value=array[index];
							feLog("        %d/%d \"%s\"\n",
									index,valueCount,value.GetText());
						}
#endif
						spec.set(propName,"string");
					}
				}
				else if(typeName=="VtArray<int>")
				{
					VtArray<int> array;
					if(attr.Get(&array,timeCode))
					{
						valueCount=array.size();
#if FE_SAU_DEBUG
						feLog("      size %d \n",valueCount);
						for(I32 index=0;index<valueCount && index<16;index++)
						{
							const int value=array[index];
							feLog("        %d/%d %d\n",
									index,valueCount,value);
						}
#endif

						if(propName=="faceVertexIndices")
						{
							vertexCount=valueCount;
						}
						else if(propName=="faceVertexCounts")
						{
							primitiveCount=valueCount;
						}
						else
						{
							spec.set(propName,"integer");
						}
					}
				}
				else if(typeName=="VtArray<float>")
				{
					VtArray<float> array;
					if(attr.Get(&array,timeCode))
					{
						valueCount=array.size();
#if FE_SAU_DEBUG
						feLog("      size %d \n",valueCount);
						for(I32 index=0;index<valueCount && index<16;index++)
						{
							const float value=array[index];
							feLog("        %d/%d %.6G\n",
									index,valueCount,value);
						}
#endif
						spec.set(propName,"real");
					}
				}
				else if(typeName=="VtArray<GfVec3f>")
				{
					VtArray<GfVec3f> array;
					if(attr.Get(&array,timeCode))
					{
						valueCount=array.size();
#if FE_SAU_DEBUG
						feLog("      size %d \n",valueCount);
						for(I32 index=0;index<valueCount && index<16;index++)
						{
							const GfVec3f& rVec=array[index];
							feLog("        %d/%d %.6G %.6G %.6G\n",
									index,valueCount,rVec[0],rVec[1],rVec[2]);
						}
#endif

						if(propName=="points")
						{
							pointCount=valueCount;
						}
						else if(valueCount==2 && propName=="extent")
						{
							const GfVec3f& rVecMin=array[0];
							const GfVec3f& rVecMax=array[1];

							const SpatialVector extentMin(
									rVecMin[0],rVecMin[1],rVecMin[2]);
							const SpatialVector extentMax(
									rVecMax[0],rVecMax[1],rVecMax[2]);

							spPrimNode->setExtentMin(extentMin);
							spPrimNode->setExtentMax(extentMax);
						}

						spec.set(propName,"vector3");
					}
				}
				else if(typeName=="VtArray<GfVec2f>")
				{
					VtArray<GfVec2f> array;
					if(attr.Get(&array,timeCode))
					{
						valueCount=array.size();
#if FE_SAU_DEBUG
						feLog("      size %d \n",valueCount);
						for(I32 index=0;index<valueCount && index<16;index++)
						{
							const GfVec2f& rVec=array[index];
							feLog("        %d/%d %.6G %.6G\n",
									index,valueCount,rVec[0],rVec[1]);
						}
#endif


						spec.set(propName,"vector3");
					}
				}

				if(isMesh && !spec.name().empty())
				{
					if(!valueCount)
					{
						rDetailSpecs.push_back(spec);
					}
					else if(isPrimVar && (valueCount==1 ||
							(primitiveCount &&
							valueCount==primitiveCount)))
					{
						rPrimitiveSpecs.push_back(spec);
					}
					else if(pointCount && valueCount==pointCount)
					{
						rPointSpecs.push_back(spec);
					}
					else if(vertexCount && valueCount==vertexCount)
					{
						rVertexSpecs.push_back(spec);
					}
				}

				if(valueCount)
				{
					line.cat(" [%d]",valueCount);
				}

				if(timeCount>1)
				{
					line.cat(" x%d",timeCount);
				}
			}
			else if(UsdRelationship rel=rProperty.As<UsdRelationship>())
			{
#if FE_SAU_DEBUG
				feLog("relationship\n");
#endif
				line+=" relationship";
			}
			else
			{
#if FE_SAU_DEBUG
				feLog("unreadable\n");
#endif
				line+=" UNREADABLE";
			}

			if(outline)
			{
				outlineAppend(line);
			}
		}
	}

	return TRUE;
}

BWORD SurfaceAccessibleUsd::save(String a_filename,sp<Catalog> a_spSettings)
{
#if FE_SAU_DEBUG
	feLog("SurfaceAccessibleUsd::save \"%s\"\n",a_filename.c_str());
#endif

	//* TODO

	return FALSE;
}

Protectable* SurfaceAccessibleUsd::clone(Protectable* pInstance)
{
#if FE_SAU_DEBUG
	feLog("SurfaceAccessibleUsd::clone\n");
#endif

	SurfaceAccessibleUsd* pSurfaceAccessibleUsd= pInstance?
			dynamic_cast<SurfaceAccessibleUsd*>(pInstance):
			new SurfaceAccessibleUsd();

	pSurfaceAccessibleUsd->setLibrary(library());

	//* TODO duplicate UsdStage

	return pSurfaceAccessibleUsd;
}

void SurfaceAccessibleUsd::attributeSpecs(
	Array<SurfaceAccessibleI::Spec>& a_rSpecs,
	String a_nodeName,
	SurfaceAccessibleI::Element a_element) const
{
	a_rSpecs.clear();

	SurfaceAccessibleI::Spec spec;

	if(a_nodeName.empty())
	{
		switch(a_element)
		{
			case SurfaceAccessibleI::e_point:
				spec.set("P","vector3");
				a_rSpecs.push_back(spec);
				break;
			case SurfaceAccessibleI::e_vertex:
				break;
			case SurfaceAccessibleI::e_primitive:
			{
				spec.set("type","string");
				a_rSpecs.push_back(spec);

				spec.set("name","string");
				a_rSpecs.push_back(spec);

				spec.set("parent","string");
				a_rSpecs.push_back(spec);

				spec.set("extentMin","vector3");
				a_rSpecs.push_back(spec);

				spec.set("extentMax","vector3");
				a_rSpecs.push_back(spec);
			}
				break;
			case SurfaceAccessibleI::e_detail:
				break;
			default:
				;
		}
		return;
	}

	try
	{
		const sp<PrimNode> spPrimNode=m_nodeMap.at(a_nodeName);

		switch(a_element)
		{
			case SurfaceAccessibleI::e_point:
				a_rSpecs=spPrimNode->pointSpecs();
				break;
			case SurfaceAccessibleI::e_vertex:
				a_rSpecs=spPrimNode->vertexSpecs();
				break;
			case SurfaceAccessibleI::e_primitive:
				a_rSpecs=spPrimNode->primitiveSpecs();
				break;
			case SurfaceAccessibleI::e_detail:
				a_rSpecs=spPrimNode->detailSpecs();
				break;
			default:
				a_rSpecs.clear();
				;
		}
	}
	catch(...)
	{
		a_rSpecs.clear();
	}
}

sp<SurfaceI> SurfaceAccessibleUsd::surface(String a_group)
{
#if FALSE
	//* TEMP just a sphere
	sp<SurfaceSphere> spSurfaceSphere=registry()->create("*.SurfaceSphere");
	if(spSurfaceSphere.isValid())
	{
		spSurfaceSphere->setCenter(SpatialVector(0.0,0.0,0.0));
		spSurfaceSphere->setRadius(1.0);
	}

	return spSurfaceSphere;
#endif

	if(a_group.empty())
	{
		return SurfaceAccessibleBase::surface(a_group);
	}

	//* NOTE copied from SurfaceAccessibleBase::surface()

	//* TODO select form
	sp<SurfaceI> spSurface;

	FEASSERT(registry().isValid());

	if(registry().isValid())
	{
		const String nodeName=a_group;

		sp<SurfaceAccessorI> spSurfacePoints=
				accessor(nodeName,e_point,e_position,e_refuseMissing);
		const I32 pointCount=spSurfacePoints.isValid()?
				spSurfacePoints->count(): 0;

		sp<SurfaceAccessorI> spSurfaceVertices=
				accessor(nodeName,e_primitive,e_vertices,e_refuseMissing);
		const I32 primitiveCount=spSurfaceVertices.isValid()?
				spSurfaceVertices->count(): 0;

		const BWORD pointCloud=(pointCount && !primitiveCount);

		sp<SurfaceAccessorI> spSurfaceProperties=
				accessor(nodeName,e_primitive,e_properties,e_refuseMissing);

		const BWORD openCurve=(spSurfaceProperties.isValid() &&
				spSurfaceProperties->count() &&
				spSurfaceProperties->integer(0,
				SurfaceAccessibleI::e_openCurve));

		if(pointCloud || openCurve)
		{
			spSurface=registry()->create(
					"SurfaceI.*.*.SurfaceCurvesAccessible");

			sp<SurfaceCurvesAccessible> spSurfaceCurvesAccessible=
					spSurface;
			if(spSurfaceCurvesAccessible.isValid())
			{
				spSurfaceCurvesAccessible->bind(
						sp<SurfaceAccessibleI>(this));
			}
		}
		else
		{
			spSurface=registry()->create(
					"SurfaceI.*.*.SurfaceTrianglesAccessible");

			sp<SurfaceTrianglesAccessible> spSurfaceTrianglesAccessible=
					spSurface;
			if(spSurfaceTrianglesAccessible.isValid())
			{
				spSurfaceTrianglesAccessible->bind(
						sp<SurfaceAccessibleI>(this));
			}
		}
	}
	else
	{
		feLog("SurfaceAccessibleUsd::surface registry is NULL\n");
	}

	if(spSurface.isNull())
	{
		feLog("SurfaceAccessibleUsd::surface failed to access surface\n");
	}

	if(spSurface.isValid())
	{
		spSurface->setNodeName(a_group);
		spSurface->setName(a_group);
	}

	return spSurface;
}

} // namespace
