/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include <usd/usd.pmh>

#define FE_SAU_DEBUG	FALSE

namespace fe
{
I32 SurfaceAccessorUsd::elementCount(
	SurfaceAccessibleI::Element a_element) const
{
	//* TODO

	switch(a_element)
	{
		case SurfaceAccessibleI::e_point:
			//* TODO check type
			return m_vec3Array.size();
		case SurfaceAccessibleI::e_vertex:
			return 0;
		case SurfaceAccessibleI::e_primitive:
			//* TODO check type
			return m_primTable.size();
		case SurfaceAccessibleI::e_detail:
			return 1;
		default:
			;
	}

	return 0;
}

void SurfaceAccessorUsd::addString(void)
{
#if FE_SAU_DEBUG
	feLog("SurfaceAccessorUsd::addString \"%s\"\n",
			m_attrName.c_str());
#endif

	//* TODO
}

void SurfaceAccessorUsd::addInteger(void)
{
#if FE_SAU_DEBUG
	feLog("SurfaceAccessorUsd::addInteger \"%s\"\n",
			m_attrName.c_str());
#endif

	//* TODO
}

void SurfaceAccessorUsd::addReal(void)
{
#if FE_SAU_DEBUG
	feLog("SurfaceAccessorUsd::addReal \"%s\"\n",
			m_attrName.c_str());
#endif

	//* TODO
}

void SurfaceAccessorUsd::addSpatialVector(void)
{
#if FE_SAU_DEBUG
	feLog("SurfaceAccessorUsd::addSpatialVector \"%s\"\n",
			m_attrName.c_str());
#endif

	//* TODO
}

BWORD SurfaceAccessorUsd::bindInternal(SurfaceAccessibleI::Element a_element,
		const String& a_name)
{
	//* TEMP
	if(a_name=="instance")
	{
		return FALSE;
	}

	const String elementName=
			SurfaceAccessibleBase::elementLayout(a_element);

#if FE_SAU_DEBUG
	feLog("SurfaceAccessorUsd::bindInternal %s \"%s\" on \"%s\"\n",
			elementName.c_str(),a_name.c_str(),m_usdPrim.GetName().GetText());
#endif

	m_element=SurfaceAccessibleI::e_point;
	if(a_element<0 && a_element>5)
	{
		a_element=SurfaceAccessibleI::e_point;
	}

	m_element=a_element;
	m_attrName=a_name;

	String usdAttrName=m_attrName;
	if(usdAttrName=="P")
	{
		usdAttrName="points";
	}
	else if(usdAttrName=="N")
	{
		usdAttrName="normals";

		//* TODO
		return FALSE;
	}

	if(m_element==SurfaceAccessibleI::e_primitive && usdAttrName=="Cd")
	{
		if(m_usdPrim.HasAttribute(TfToken("primvars:Cd")))
		{
			usdAttrName="primvars:Cd";
		}
		else if(m_usdPrim.HasAttribute(TfToken("primvars:color")))
		{
			usdAttrName="primvars:color";
		}
		else if(m_usdPrim.HasAttribute(TfToken("primvars:displayColor")))
		{
			usdAttrName="primvars:displayColor";
		}
	}

	if(m_element==SurfaceAccessibleI::e_vertex && usdAttrName=="uv")
	{
		usdAttrName="primvars:uv";
	}

	m_transform=getTransform(m_usdPrim);

	std::vector<UsdProperty> propArray=m_usdPrim.GetProperties();
	const I32 propCount=propArray.size();
	for(I32 propIndex=0;propIndex<propCount;propIndex++)
	{
		UsdProperty& rProperty=propArray[propIndex];

		const String attrName=rProperty.GetName().GetText();

		if(m_attribute==SurfaceAccessibleI::e_vertices &&
				(attrName=="faceVertexCounts" || attrName=="faceVertexIndices"))
		{
		}
		else if(attrName!=usdAttrName)
		{
			continue;
		}

		if(UsdAttribute attr=rProperty.As<UsdAttribute>())
		{
			std::vector<double> timeArray;
			attr.GetTimeSamples(&timeArray);

			const I32 timeCount=timeArray.size();

			const UsdTimeCode timeCode=timeCount?
					UsdTimeCode(m_frame): UsdTimeCode::Default();

			if(m_attribute==SurfaceAccessibleI::e_vertices &&
					attrName=="faceVertexCounts")
			{
				VtArray<int> intArray;
				if(!attr.Get(&intArray,timeCode))
				{
					continue;
				}
				const I32 primitiveCount=intArray.size();
				m_primTable.resize(primitiveCount);

				I32 firstVertex=0;
				for(I32 primitiveIndex=0;primitiveIndex<primitiveCount;
						primitiveIndex++)
				{
					const I32 vertCount=intArray[primitiveIndex];
					fe::set(m_primTable[primitiveIndex],firstVertex,vertCount);
					firstVertex+=vertCount;
				}
				continue;
			}

			if(m_attribute==SurfaceAccessibleI::e_vertices &&
					attrName=="faceVertexIndices")
			{
				if(!attr.Get(&m_intArray,timeCode))
				{
					feLog("SurfaceAccessorUsd::bindInternal"
							" failed to access vertex array\n");
				}
				continue;
			}

//~			if(attrName!=usdAttrName)
//~			{
//~				continue;
//~			}

			const String typeName=
					attr.GetTypeName().GetType().GetTypeName().c_str();
			const String roleName=attr.GetRoleName().GetText();

#if FE_SAU_DEBUG
			feLog("  attr \"%s\" type \"%s\" role \"%s\"\n",
					attrName.c_str(),typeName.c_str(),roleName.c_str());
#endif

			if(typeName=="TfToken")
			{
				TfToken value;
				if(attr.Get(&value,timeCode))
				{
					m_tokenArray.resize(1);
					m_tokenArray[0]=value;
					return TRUE;
				}
#if FE_SAU_DEBUG
				else
				{
					feLog("SurfaceAccessorUsd::bindInternal"
							" failed to access TfToken %s \"%s\"\n",
							elementName.c_str(),m_attrName.c_str());
				}
#endif
			}
			else if(typeName=="bool")
			{
				bool value;
				if(attr.Get(&value,timeCode))
				{
					m_boolArray.resize(1);
					m_boolArray[0]=value;
					return TRUE;
				}
#if FE_SAU_DEBUG
				else
				{
					feLog("SurfaceAccessorUsd::bindInternal"
							" failed to access bool %s \"%s\"\n",
							elementName.c_str(),m_attrName.c_str());
				}
#endif
			}
			else if(typeName=="double")
			{
				double value;
				if(attr.Get(&value,timeCode))
				{
					m_doubleArray.resize(1);
					m_doubleArray[0]=value;
					return TRUE;
				}
#if FE_SAU_DEBUG
				else
				{
					feLog("SurfaceAccessorUsd::bindInternal"
							" failed to access double %s \"%s\"\n",
							elementName.c_str(),m_attrName.c_str());
				}
#endif
			}
			else if(typeName=="VtArray<TfToken>")
			{
				if(attr.Get(&m_tokenArray,timeCode))
				{
					return TRUE;
				}
#if FE_SAU_DEBUG
				else
				{
					feLog("SurfaceAccessorUsd::bindInternal"
							" failed to access TfToken array %s \"%s\"\n",
							elementName.c_str(),m_attrName.c_str());
				}
#endif
			}
			else if(typeName=="VtArray<int>")
			{
				if(attr.Get(&m_intArray,timeCode))
				{
					return TRUE;
				}
#if FE_SAU_DEBUG
				else
				{
					feLog("SurfaceAccessorUsd::bindInternal"
							" failed to access int array %s \"%s\"\n",
							elementName.c_str(),m_attrName.c_str());
				}
#endif
			}
			else if(typeName=="VtArray<float>")
			{
				if(attr.Get(&m_floatArray,timeCode))
				{
					return TRUE;
				}
#if FE_SAU_DEBUG
				else
				{
					feLog("SurfaceAccessorUsd::bindInternal"
							" failed to access float array %s \"%s\"\n",
							elementName.c_str(),m_attrName.c_str());
				}
#endif
			}
			else if(typeName=="VtArray<GfVec3f>")
			{
				if(attr.Get(&m_vec3Array,timeCode))
				{
					return TRUE;
				}
#if FE_SAU_DEBUG
				else
				{
					feLog("SurfaceAccessorUsd::bindInternal"
							" failed to access GfVec3f array %s \"%s\"\n",
							elementName.c_str(),m_attrName.c_str());
				}
#endif
			}
			else if(typeName=="VtArray<GfVec2f>")
			{
				if(attr.Get(&m_vec2Array,timeCode))
				{
					return TRUE;
				}
#if FE_SAU_DEBUG
				else
				{
					feLog("SurfaceAccessorUsd::bindInternal"
							" failed to access GfVec2f array %s \"%s\"\n",
							elementName.c_str(),m_attrName.c_str());
				}
#endif
			}
		}
	}

	if(m_intArray.size() && m_primTable.size())
	{
#if FE_SAU_DEBUG
		feLog("SurfaceAccessorUsd::bindInternal vertices %d primitives %d\n",
				m_intArray.size(),m_primTable.size());
#endif
		return TRUE;
	}

#if FE_SAU_DEBUG
	feLog("SurfaceAccessorUsd::bindInternal %s \"%s\" not found\n",
			elementName.c_str(),a_name.c_str());
#endif

	return FALSE;
}

SpatialTransform SurfaceAccessorUsd::getTransform(UsdPrim a_usdPrim)
{
	SpatialTransform xform;
	setIdentity(xform);

	//* NOTE direct UsdPrim::GetProperty() doesn't seem to work

	std::vector<UsdProperty> propArray=a_usdPrim.GetProperties();
	const I32 propCount=propArray.size();
	for(I32 propIndex=0;propIndex<propCount;propIndex++)
	{
		UsdProperty& rProperty=propArray[propIndex];

		const String attrName=rProperty.GetName().GetText();
		if(attrName!="xformOp:transform")
		{
			continue;
		}

		if(UsdAttribute attr=rProperty.As<UsdAttribute>())
		{
			std::vector<double> timeArray;
			attr.GetTimeSamples(&timeArray);

			const I32 timeCount=timeArray.size();

			const UsdTimeCode timeCode=timeCount?
					UsdTimeCode(m_frame): UsdTimeCode::Default();

			GfMatrix4d matrix;
			if(attr.Get(&matrix,timeCode))
			{
				fe::set(xform,matrix.GetArray());
			}
			else
			{
				feLog("SurfaceAccessorUsd::getTransform"
						" failed to access transform\n");
			}

			break;
		}
	}

	UsdPrim primParent=a_usdPrim.GetParent();
	if(primParent)
	{
		//* NOTE recursive
		xform=xform*getTransform(primParent);
	}

	return xform;
}


} // namespace
