/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __usd_SurfaceAccessorUsd_h__
#define __usd_SurfaceAccessorUsd_h__

namespace fe
{

/**************************************************************************//**
    @brief USD Surface Editing Implementation

	@ingroup usd
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessorUsd:
	public SurfaceAccessorBase
{
	public:
						SurfaceAccessorUsd(void)
						{	setName("SurfaceAccessorUsd"); }
virtual					~SurfaceAccessorUsd(void)							{}

						using SurfaceAccessorBase::set;
						using SurfaceAccessorBase::append;
						using SurfaceAccessorBase::spatialVector;

						//* as SurfaceAccessorI
		BWORD			bind(SurfaceAccessibleI::Element a_element,
								SurfaceAccessibleI::Attribute a_attribute)
						{
							m_attribute=a_attribute;

							String name;
							switch(a_attribute)
							{
								case SurfaceAccessibleI::e_generic:
								case SurfaceAccessibleI::e_position:
									name="P";
									break;
								case SurfaceAccessibleI::e_normal:
									name="N";
									break;
								case SurfaceAccessibleI::e_uv:
									name="uv";
									break;
								case SurfaceAccessibleI::e_color:
									name="Cd";
									break;
								case SurfaceAccessibleI::e_vertices:
									name="vertices";
									FEASSERT(a_element==
											SurfaceAccessibleI::e_primitive);
									break;
								case SurfaceAccessibleI::e_properties:
									name="properties";
									FEASSERT(a_element==
											SurfaceAccessibleI::e_primitive);
									break;
							}
							return bindInternal(a_element,name);
						}
		BWORD			bind(SurfaceAccessibleI::Element a_element,
								const String& a_name)
						{
							m_attribute=SurfaceAccessibleI::e_generic;
							return bindInternal(a_element,a_name);
						}
virtual	U32				count(void) const
						{
							if(m_attribute!=SurfaceAccessibleI::e_generic &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									m_attribute!=
									SurfaceAccessibleI::e_properties &&
									!isBound())
							{
								return 0;
							}

							return elementCount(m_element);
						}
virtual	U32				subCount(U32 a_index) const
						{
							if(m_attribute!=SurfaceAccessibleI::e_generic &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									m_attribute!=
									SurfaceAccessibleI::e_properties &&
									!isBound())
							{
								return 0;
							}

							if(m_attribute==SurfaceAccessibleI::e_vertices &&
									a_index<m_primTable.size())
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitive);

								return m_primTable[a_index][1];
							}

							return 1;
						}

virtual	void			set(U32 a_index,U32 a_subIndex,String a_string)
						{
							if(m_element!=SurfaceAccessibleI::e_pointGroup &&
									m_element!=
									SurfaceAccessibleI::e_primitiveGroup &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addString();
							}

							if(!isBound())
							{
								return;
							}

							//* TODO
						}

virtual	String			string(U32 a_index,U32 a_subIndex=0)
						{
							if(!isBound())
							{
								return 0;
							}

							const I32 index=
									(m_tokenArray.size()==1)? 0: a_index;

							if(a_index<m_tokenArray.size())
							{
								return m_tokenArray[a_index].GetText();
							}

							return String();
						}

virtual	void			set(U32 a_index,U32 a_subIndex,I32 a_integer)
						{
							if(m_element!=SurfaceAccessibleI::e_pointGroup &&
									m_element!=
									SurfaceAccessibleI::e_primitiveGroup &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addInteger();
							}

							if(!isBound())
							{
								return;
							}

							//* TODO
						}

virtual	I32				integer(U32 a_index,U32 a_subIndex=0)
						{
							if(!isBound())
							{
								return 0;
							}

							if(m_element==SurfaceAccessibleI::e_primitive &&
									m_attribute==SurfaceAccessibleI::e_vertices
									&& a_index<m_primTable.size())
							{
								const Vector2i& rEntry=m_primTable[a_index];
								if(a_subIndex<rEntry[1])
								{
									return m_intArray[rEntry[0]+a_subIndex];
								}
							}

							const I32 index=(m_intArray.size()==1 ||
									m_boolArray.size()==1)? 0: a_index;

							if(index<m_intArray.size())
							{
								return m_intArray[index];
							}

							if(index<m_boolArray.size())
							{
								return m_boolArray[index];
							}

							return 0;
						}

virtual	I32				append(SurfaceAccessibleI::Form a_form)
						{
							//* TODO

							return -1;
						}

virtual	void			append(U32 a_index,I32 a_integer)
						{
							//* TODO
						}

virtual	void			set(U32 a_index,U32 a_subIndex,Real a_real)
						{
							if(m_element!=SurfaceAccessibleI::e_pointGroup &&
									m_element!=
									SurfaceAccessibleI::e_primitiveGroup &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addReal();
							}

							if(!isBound())
							{
								return;
							}

							//* TODO
						}

virtual	Real			real(U32 a_index,U32 a_subIndex=0)
						{
							if(!isBound())
							{
								return Real(0);
							}

							const I32 index=(m_floatArray.size()==1 ||
									m_doubleArray.size()==1)? 0: a_index;

							if(index<m_floatArray.size())
							{
								return m_floatArray[index];
							}

							if(index<m_doubleArray.size())
							{
								return m_doubleArray[index];
							}

							return Real(0);
						}

virtual	void			set(U32 a_index,U32 a_subIndex,
								const SpatialVector& a_vector)
						{
							if(m_element!=SurfaceAccessibleI::e_pointGroup &&
									m_element!=
									SurfaceAccessibleI::e_primitiveGroup &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addSpatialVector();
							}

							if(!isBound())
							{
								return;
							}

							//* TODO
						}
virtual	SpatialVector	spatialVector(U32 a_index,U32 a_subIndex=0)
						{
							if(!isBound())
							{
								return SpatialVector(0.0,0.0,0.0);
							}

							const I32 index=(m_vec2Array.size()==1 ||
									m_vec3Array.size()==1)? 0: a_index;

							if(index<m_vec3Array.size())
							{
								const GfVec3f& vec3=m_vec3Array[index];
								SpatialVector result(vec3[0],vec3[1],vec3[2]);
								if(m_attrName=="P")
								{
									return transformVector(m_transform,result);
								}
								if(m_attrName=="N")
								{
									return rotateVector(m_transform,result);
								}
								return result;
							}

							if(index<m_vec2Array.size())
							{
								const GfVec2f& vec3=m_vec2Array[index];
								SpatialVector result(vec3[0],vec3[1],0.0);
								return result;
							}

							//* TODO primitive position

							return SpatialVector(0.0,0.0,0.0);
						}

						//* USD specific
		void			setUsdPrim(UsdPrim& a_usdPrim)
						{	m_usdPrim=a_usdPrim; }

		UsdPrim&		usdPrim(void)		{ return m_usdPrim; }

		void			setFrame(Real a_frame)
						{	m_frame=a_frame; }

	private:

virtual	BWORD			bindInternal(SurfaceAccessibleI::Element a_element,
								const String& a_name);

		void			addString(void);
		void			addInteger(void);
		void			addReal(void);
		void			addSpatialVector(void);

		I32				elementCount(
								SurfaceAccessibleI::Element a_element) const;

		BWORD			isBound(void) const
						{	return m_usdPrim.IsValid(); }

		SpatialTransform	getTransform(UsdPrim a_usdPrim);

		SpatialTransform	m_transform;
		Real				m_frame;

		UsdPrim				m_usdPrim;
		UsdAttribute		m_usdAttribute;

		Array<Vector2i>		m_primTable;	//* first vertex, subCount

		VtArray<TfToken>	m_tokenArray;
		VtArray<bool>		m_boolArray;
		VtArray<int>		m_intArray;
		VtArray<float>		m_floatArray;
		VtArray<double>		m_doubleArray;
		VtArray<GfVec3f>	m_vec3Array;
		VtArray<GfVec2f>	m_vec2Array;
};

} // namespace

#endif /* __usd_SurfaceAccessorUsd_h__ */
