/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include <usd/usd.pmh>

#define FE_SAUG_DEBUG	FALSE

namespace fe
{
I32 SurfaceAccessorUsdGraph::elementCount(
	SurfaceAccessibleI::Element a_element) const
{
	switch(a_element)
	{
		case SurfaceAccessibleI::e_point:
			return 8*m_pNodeArray->size();
		case SurfaceAccessibleI::e_vertex:
			return 0;
		case SurfaceAccessibleI::e_primitive:
			return m_pNodeArray->size();
		case SurfaceAccessibleI::e_detail:
			return 0;
		default:
			;
	}

	return 0;
}

SpatialVector SurfaceAccessorUsdGraph::extentCorner(
	I32 a_primitiveIndex,I32 a_offset)
{
	if(a_primitiveIndex<m_pNodeArray->size())
	{
		const sp<SurfaceAccessibleUsd::PrimNode>& rspPrimNode=
				(*m_pNodeArray)[a_primitiveIndex];
		const SpatialVector extentMin=rspPrimNode->extentMin();
		const SpatialVector extentMax=rspPrimNode->extentMax();

		switch(a_offset)
		{
			case 0:
				return extentMin;
			case 1:
				return SpatialVector(extentMax[0],extentMin[1],extentMin[2]);
			case 2:
				return SpatialVector(extentMax[0],extentMin[1],extentMax[2]);
			case 3:
				return SpatialVector(extentMin[0],extentMin[1],extentMax[2]);
			case 4:
				return extentMax;
			case 5:
				return SpatialVector(extentMin[0],extentMax[1],extentMax[2]);
			case 6:
				return SpatialVector(extentMin[0],extentMax[1],extentMin[2]);
			case 7:
				return SpatialVector(extentMax[0],extentMax[1],extentMin[2]);
			default:
				;
		}
	}

	return SpatialVector(0.0,0.0,0.0);
}

BWORD SurfaceAccessorUsdGraph::bindInternal(
		SurfaceAccessibleI::Element a_element,
		const String& a_name)
{
	if(a_element!=SurfaceAccessibleI::e_primitive &&
			!(a_element==SurfaceAccessibleI::e_point && a_name=="P"))
	{
		return FALSE;
	}

	//* TODO
	if(a_name=="instance")
	{
		return FALSE;
	}

	const String elementName=
			SurfaceAccessibleBase::elementLayout(a_element);

#if FE_SAUG_DEBUG
	feLog("SurfaceAccessorUsdGraph::bindInternal %s \"%s\"\n",
			elementName.c_str(),a_name.c_str());
#endif

	m_element=a_element;
	m_attrName=a_name;

	sp<SurfaceAccessibleUsd> spSurfaceAccessibleUsd=m_spSurfaceAccessibleI;
	if(spSurfaceAccessibleUsd.isNull())
	{
		return FALSE;
	}

	m_pNodeArray=&spSurfaceAccessibleUsd->nodeArray();
	const I32 nodeCount=m_pNodeArray? m_pNodeArray->size(): 0;

#if FE_SAUG_DEBUG
	feLog("SurfaceAccessorUsdGraph::bindInternal nodeCount %d\n",nodeCount);
#endif

	if(!nodeCount)
	{
#if FE_SAUG_DEBUG
		feLog("SurfaceAccessorUsdGraph::bindInternal %s \"%s\" not found\n",
				elementName.c_str(),a_name.c_str());
#endif

		return FALSE;
	}

	return TRUE;
}

} // namespace
