import sys
import os
import re
forge = sys.modules["forge"]

def setup(module):

	version_paths = [	"imaging/hd",
						"imaging/hdx",
						"usdImaging/usdImaging",
						"usdImaging/usdImagingGL" ]

	if forge.usd_imaging_install == "":
		forge.usd_imaging_install = forge.usd_install

	delim = " "
	for path in version_paths:
		version = ""
		for line in open(forge.usd_imaging_install + "/include/pxr/" + path + "/version.h").readlines():
			if re.match("#define .*_API_VERSION", line):
				version = line.split()[2].rstrip()
		if version != "":
			sys.stdout.write(delim + version)
		delim = "/"

	deplibs =	forge.corelibs

	module.cppmap = {}
	module.cppmap["std"] = "-std=c++11"
	module.cppmap["usd"] = "-D BUILD_COMPONENT_SRC_PREFIX=\\\"pxr/\\\""
	module.cppmap["usd"] += " -D BUILD_OPTLEVEL_OPT"

	module.includemap = {}

	module.includemap["python"] = "/usr/include/python2.7"

	module.includemap["tbb"] = "/opt/hfs15.0.416/toolkit/include"

	module.includemap["exr"] = module.modPath + "/OpenEXR/include/OpenEXR"
#	module.includemap["exr"] = forge.openexr_source + "/include/OpenEXR"

	module.includemap["usd"] = forge.usd_install + "/include"

	usd_install_lib = forge.usd_install + "/lib"

	srcList = [	"hydra.pmh",
				"DrawHydra",
				"HydraDelegate",
				"HydraDriver",
				"hydraDL" ]

	dll = module.DLL( "fexHydraDL", srcList )
	dll.linkmap = {}
	dll.linkmap["usd"] = "-Xlinker -rpath='" + usd_install_lib + "'"
	dll.linkmap["usd"] += " -L " + usd_install_lib

	if forge.usd_imaging_install != forge.usd_install:
		module.includemap["usd_imaging"] = forge.usd_imaging_install + "/include"

		usd_imaging_install_lib = forge.usd_imaging_install + "/lib"
		dll.linkmap["usd"] += " -Xlinker -rpath='" + usd_imaging_install_lib + "'"
		dll.linkmap["usd"] += " -L " + usd_imaging_install_lib

	dll.linkmap["usd"] += " -l" + forge.usd_imaging_prefix + "hd" + forge.usd_imaging_postfix
	dll.linkmap["usd"] += " -l" + forge.usd_imaging_prefix + "hdx" + forge.usd_imaging_postfix
	dll.linkmap["gfxlibs"] = forge.gfxlibs

	#HACK for OIIO dependency
#	dll.linkmap["usd"] += " -lssl"

#	dll.linkmap["glew"] = "-lGLEW"
#	dll.linkmap["python"] = "-lpython2.7"

	srcList = [	"usd.pmh",
				"SurfaceAccessibleUsd",
				"SurfaceAccessorUsd",
				"SurfaceAccessorUsdGraph",
				"usdDL" ]

	dll = module.DLL( "fexUsdDL", srcList )
	dll.linkmap = {}
	dll.linkmap["usd"] = "-Xlinker -rpath='" + usd_install_lib + "'"
	dll.linkmap["usd"] += " -L " + usd_install_lib
	dll.linkmap["usd"] += " -l" + forge.usd_prefix + "usd" + forge.usd_postfix

	forge.tests += [
		("inspect.exe",		"fexUsdDL",				None,		None) ]

def auto(module):
	test_file = """
#include "pxr/imaging/hd/version.h"

int main(void)
{
	return(0);
}
    \n"""

	includemap = { "usd": forge.usd_install + "/include" }

	if forge.usd_imaging_install != forge.usd_install:
		includemap["usd_imaging"] = forge.usd_imaging_install + "/include"

	linkmap = {}

	return forge.cctest(test_file, linkmap, includemap)
