/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include <vdb/vdb.pmh>

#define FE_SAV_DEBUG			FALSE

namespace fe
{

BWORD SurfaceAccessibleVDB::load(String a_filename,sp<Catalog> a_spSettings)
{
#if FE_SAV_DEBUG
	feLog("SurfaceAccessibleVDB::load \"%s\"\n",a_filename.c_str());
#endif

	std::string filename(a_filename.c_str());
	openvdb::io::File vdbFile(filename);

	try
	{
		if(!vdbFile.open())
		{
			feLog("SurfaceAccessibleVDB::load open file failed\n");
			return FALSE;
		}
	}
	catch(const std::exception& std_e)
	{
		feLog("SurfaceAccessibleVDB::load exception:\n  \"%s\"\n",
				std_e.what());
		return FALSE;
	}

#if FE_SAV_DEBUG
	feLog("SurfaceAccessibleVDB::load file opened\n");
#endif

	openvdb::Name name;

	openvdb::io::File::NameIterator it=vdbFile.beginName();
	while(it!=vdbFile.endName())
	{
		name= *it;
		++it;

#if FE_SAV_DEBUG
		feLog("SurfaceAccessibleVDB::load found grid \"%s\"\n",name.c_str());
#endif
	}

	m_gridPtr=vdbFile.readGrid(name);

	return bool(m_gridPtr);
}

sp<SurfaceI> SurfaceAccessibleVDB::surface(String a_group)
{
#if FE_SAV_DEBUG
	feLog("SurfaceAccessibleVDB::surface\n");
#endif

	sp<SurfaceVDB> spSurfaceVDB=registry()->create("*.SurfaceVDB");
	if(spSurfaceVDB.isValid())
	{
		//* TODO better
		void* pPtr=&m_gridPtr;

		sp<TypeMaster> spTypeMaster=registry()->master()->typeMaster();
		Instance instance;
		instance.set(spTypeMaster,pPtr);
		spSurfaceVDB->bind(instance);
	}

#if FE_SAV_DEBUG
	feLog("SurfaceAccessibleVDB::surface valid %d\n",spSurfaceVDB.isValid());
#endif

	return spSurfaceVDB;
}

} // namespace
